"""
bento_meta.util.cypher.entities

Representations of cypher nodes, relationships, properties, paths
"""
import re
from pdb import set_trace
from bento_meta.util.cypher.functions import Func
from copy import deepcopy as clone


def countmaker(max=10000):
    return (x for x in range(max))


class Entity(object):
    """A property graph Entity. Base class."""
    def __init__(self):
        self.As = None
        ct = None
        try:
            ct = next(type(self).count)
        except StopIteration:
            type(self)._reset_counter()
            ct = next(type(self).count)
        self.var = type(self).__name__[0].lower() + str(ct)

    @classmethod
    def _reset_counter(cls):
        cls.count = countmaker()

    def pattern(self):
        """Render entity as a match pattern."""
        pass

    def condition(self):
        """Render entity as a condition (for WHERE, e.g.)."""
        pass

    def Return(self):
        """Render entity as a return value."""
        pass

    def _add_props(self, props):
        if not type(self) == N and not type(self) == R:
            return False
        if not props:
            return True
        if type(props) == P:
            props.entity = self
            self.props[props.handle] = props
        elif type(props) == list:
            for p in props:
                p.entity = self
            for p in props:
                self.props[p.handle] = p
        elif type(props) == dict:
            for hdl in props:
                self.props[hdl] = P(handle=hdl, value=props[hdl])
                self.props[hdl].entity = self
        else:
            raise RuntimeError("Can't interpret props arg '{}'".format(props))
        return True


class N(Entity):
    """A property graph Node."""
    count = countmaker()

    def __init__(self, label=None, props=None, As=None):
        super().__init__()
        self.props = {}
        self.label = label
        self._add_props(props)
        self.As = As

    def relate_to(self, r, n):
        # always obj --> arg direction
        return r.relate(self, n)

    def pattern(self):
        ret = ",".join([p.pattern() for p in
                        self.props.values() if p.pattern()])
        if (len(ret)):
            ret = " {"+ret+"}"
        if self.label:
            ret = "({}:{}{})".format(self.var, self.label, ret)
        else:
            ret = "({}{})".format(self.var, ret)
        return ret

    def condition(self):
        return [p.condition() for p in self.props.values()]

    def Return(self):
        ret = " as {}".format(self.As) if self.As else ""
        ret = "{}{}".format(self.var, ret)
        return ret


class R(Entity):
    """A property graph Relationship or edge."""
    count = countmaker()

    def __init__(self, Type=None, props=None, As=None, _dir='_right'):
        super().__init__()
        self.props = {}
        self.Type = Type
        self._add_props(props)
        self._join = []
        self._dir = _dir
        self.As = As

    # n --> m direction
    def relate(self, n, m):
        return T(n, self, m) if self._dir == '_right' else T(m, self, n)

    def pattern(self):
        ret = ",".join([p.pattern() for p in
                        self.props.values() if p.pattern()])
        if (len(ret)):
            ret = " {"+ret+"}"
        if self.Type:
            ret = "-[{}:{}{}]-".format(self.var, self.Type, ret)
        else:
            ret = "-[{}{}]-".format(self.var, ret)
        return ret

    def condition(self):
        return [p.condition() for p in self.props.values()]

    def Return(self):
        ret = " as {}".format(self.As) if self.As else ""
        ret = "{}{}".format(self.var, ret)
        return ret

class VarLenR(R):
    """Variable length property graph Relationship or edge."""
    def __init__(self,
        min_len: int = -1,
        max_len: int = -1, 
        Type=None, props=None, As=None, _dir='_right'):
        super().__init__()
        self.props = {}
        self.Type = Type
        self._add_props(props)
        self.min_len = min_len
        self.max_len = max_len

    def pattern(self):
        ret = ",".join([p.pattern() for p in
                        self.props.values() if p.pattern()])
        var_len = "*"
        if self.min_len < 0:
            min_len_str = ""
        else:
            min_len_str = str(self.min_len)
        if self.max_len < 0:
            max_len_str = ""
        else:
            max_len_str = str(self.max_len)
        if min_len_str or max_len_str:
            var_len += f"{min_len_str}..{max_len_str}"
        if len(ret):
            ret = " {"+ret+"}"
        if self.Type:
            ret = f"-[:{self.Type}{ret}{var_len}]-"
        else:
            ret = f"-[{ret}{var_len}]-"
        return ret

class N0(N):
    """Completely anonymous node ()."""
    def __init__(self):
        super().__init__()
        self.var = None

    def pattern(self):
        return "()"

    def Return(self):
        return None


class R0(R):
    """Completely anonymous relationship -[]-, i.e. --"""
    def __init__(self):
        super().__init__()
        self.var = None

    def pattern(self):
        return "--"

    def Return(self):
        return None


class P(Entity):
    """A property graph Property."""
    count = countmaker()
    parameterize = False

    def __init__(self, handle, value=None, As=None):
        super().__init__()
        self.handle = handle
        self.value = value
        self.As = As
        self.entity = None
        self.param = {self.var: value} if value else None

    def pattern(self):
        if self.value:
            if not self.parameterize:
                if not type(self.value) == str:
                    return "{}:{}".format(self.handle, str(self.value))
                elif re.match("^\\s*[$]", self.value):  # a parameter
                    return "{}:{}".format(self.handle, self.value)
                else:
                    return "{}:'{}'".format(self.handle, self.value)
            else:
                return "{}:${}".format(self.handle, self.var)
        else:
            return None

    def condition(self):
        if self.value and self.entity:
            if not self.parameterize:
                if not type(self.value) == str:
                    return "{}.{} = {}".format(self.entity.var, self.handle,
                                               str(self.value))
                elif re.match("^\\s*[$]", self.value):  # a parameter
                    return "{}.{} = {}".format(self.entity.var, self.handle,
                                               self.value)
                else:
                    return "{}.{} = '{}'".format(self.entity.var, self.handle,
                                                 self.value)
            else:
                return "{}.{} = ${}".format(self.entity.var, self.handle,
                                            self.var)
        else:
            return None

    def Return(self):
        ret = " as {}".format(self.As) if self.As else ""
        if self.entity:
            return "{}.{}{}".format(self.entity.var, self.handle, ret)
        else:
            return None


class T(Entity):
    """A property graph Triple; i.e., (n)-[r]->(m)."""
    count = countmaker()

    def __init__(self, n, r, m):
        super().__init__()
        self._from = n
        self._to = m
        self._edge = r

    def nodes(self):
        return [self._from, self._to]

    def edge(self):
        return self._edge

    def edges(self):
        return [self.edge()]

    def pattern(self):
        return self._from.pattern()+self._edge.pattern()+">"+self._to.pattern()

    def condition(self):
        return self.pattern()

    def Return(self):
        return [x.Return() for x
                in (self._from, self._edge, self._to) if x.var]

class NoDirT(T):
    """A directionless property graph Triple; i.e., (n)-[r]-(m)."""
    def __init__(self, *args):
        super().__init__(*args)

    def pattern(self):
        return self._from.pattern()+self._edge.pattern()+self._to.pattern()

    def nodes(self):
        return [self._from, self._to]

    def edge(self):
        return self._edge

    def edges(self):
        return [self.edge()]

class G(Entity):
    """A property graph Path.
    Defined as an ordered set of partially overlapping triples."""
    count = countmaker()

    def __init__(self, *args):
        super().__init__()
        self._pattern = None
        self.triples = []
        args = list(args)
        self._create_path(args)

    def _create_path(self, args):
        numargs = len(args)
        scr = []
        while args:
            ent = args.pop(0)
            if len(scr) == 0:
                if isinstance(ent, N):
                    scr.append(ent)
                elif isinstance(ent, R):
                    if self.triples:
                        scr.append(ent)
                    else:
                        raise RuntimeError(
                            "Entity '{}' is not valid at arg position {}."
                            .format(ent, numargs-len(args))
                        )
                elif isinstance(ent, (T, G)):
                    if not self._append(ent):
                        raise RuntimeError(
                            "Adjacent triples/paths do not overlap, "
                            "at arg position {}.".format(numargs-len(args))
                        )
                    scr = []  # reset parse
                else:
                    raise RuntimeError(
                        "Entity '{}' is not valid at arg position {}."
                        .format(ent, numargs-len(args))
                    )
            elif len(scr) == 1:
                if (isinstance(scr[0], N) and isinstance(ent, R)):
                    scr.append(ent)
                elif isinstance(scr[0], R):
                    if isinstance(ent, N):
                        scr.append(ent)
                    elif isinstance(ent, (T, G)) and scr[0]._join:
                        # may be able to link self to ent with _join hints
                        r = scr[0]
                        n1 = [x for x in self.triples[-1].nodes()
                              if x.label == r._join[0]]
                        n2 = [x for x in ent.nodes() if x.label == r._join[1]]
                        if n1 and n2:
                            t = r.relate(n1[0], n2[0])
                            if not self._append(t):
                                raise RuntimeError("WTF error; this should not fail")
                            if not self._append(ent):
                                raise RuntimeError(
                                    "Found right-hand node but could not append path"
                                    )
                            scr = []  # success
                        else:
                            raise RuntimeError(
                                "Can't find endpoint nodes specified in _join for "
                                "relationship at arg position {}".format(numargs-len(args)-1)
                                )
                    else:
                        raise RuntimeError(
                            "Ends of relationship are ambiguous; relationship "
                            "at arg position {} needs _join hints".format(numargs-len(args)-1)
                            )
                else:
                    raise RuntimeError(
                        "Entity '{}' is not valid at arg position {}."
                        .format(ent, numargs-len(args))
                    )
                pass
            elif len(scr) == 2:
                if isinstance(scr[0], N):
                    success = None
                    if isinstance(ent, N):
                        success = self._append(scr[1].relate(scr[0], ent))
                    elif isinstance(ent, T):
                        success = self._append(scr[1].relate(scr[0],
                                                             ent._from))
                        if success:
                            success = self._append(ent)
                    elif isinstance(ent, G):
                        if len(ent.triples) > 1:
                            raise RuntimeError(
                                "Can't create start triple, "
                                "to-node is ambiguous, "
                                "at arg position {}.".format(numargs-len(args))
                            )
                        success = self._append(
                            scr[1].relate(scr[0], ent.triples[0]._from))
                    else:
                        raise RuntimeError(
                            "Entity '{}' is not valid at arg position {}."
                            .format(ent, numargs-len(args))
                        )
                    if not success:
                        raise RuntimeError(
                            "Resulting adjacent triples/paths do not overlap, "
                            "at arg position {}."
                            .format(numargs-len(args))
                        )
                elif isinstance(scr[0], R):
                    if not self._append(scr[0].relate(self.triples[-1]._to,
                                                      scr[1])):
                        raise RuntimeError(
                            "Resulting adjacent triples/paths do not overlap, "
                            "at arg position {}."
                            .format(numargs-len(args))
                        )
                scr = []
            else:
                raise RuntimeError("Shouldn't happen.")
        if scr:
            if len(scr) == 2 and isinstance(
                    scr[0], R) and isinstance(scr[1], N):
                if len(self.triples) > 1:
                    raise RuntimeError(
                        "Can't create end triple, from-node is ambiguous, "
                        "at arg position {}.".format(numargs-len(args))
                    )
                if not self._append(
                        scr[0].relate(self.triples[-1]._to, scr[1])):
                    raise RuntimeError(
                        "Resulting adjacent triples/paths do not overlap, "
                        "at arg position {}."
                        .format(numargs-len(args))
                    )
            else:
                raise RuntimeError("Args do not define a complete Path.")

    def _append(self, ent):
        def _overlap(s, t):
            if s == t:
                return 0b000
            if isinstance(s, G):
                s = s.triples[-1]
            if isinstance(t, G):
                t = t.triples[0]
            if s._from == t._from or (
                    s._from.var and s._from.var == t._from.var):
                return 0b001
            if s._from == t._to or (
                    s._from.var and s._from.var == t._to.var):
                return 0b010
            if s._to == t._from or (
                    s._to.var and s._to.var == t._from.var):
                return 0b011
            if s._to == t._to or (
                    s._to.var and s._to.var == t._to.var):
                return 0b100
            return -1

        if not isinstance(ent, (T, G)):
            raise RuntimeError("Can't append a {} to a Path."
                               .format(type(ent).__name__))
        if self.triples:
            ovr = _overlap(self.triples[-1], ent)
            if ovr == 0:
                return True  # same object, skip
            if ovr < 0:
                return False  # Adjacent triples/paths do not overlap
        if isinstance(ent, T):
            self.triples.append(ent)
        elif isinstance(ent, G):
            self.triples.extend(ent.triples)
        return True

    def nodes(self):
        ret = set()
        for t in self.triples:
            ret.add(t._from)
            ret.add(t._to)
        return list(ret)

    def edges(self):
        ret = set()
        for t in self.triples:
            ret.add(t._edge)
        return list(ret)

    def pattern(self):
        def jn(trps, acc, ret):
            if not trps:
                ret.append(acc)
                return ret
            trp = trps.pop(0)
            if not acc:
                return jn(trps, [trp._from, trp._edge, ">", trp._to], ret)
            else:
                if trp._from == acc[0]:
                    acc[0:0] = [trp._to, "<", trp._edge]
                elif trp._from == acc[-1]:
                    acc.extend([trp._edge, ">", trp._to])
                elif trp._to == acc[0]:
                    acc[0:0] = [trp._from, trp._edge, ">"]
                elif trp._to == acc[-1]:
                    acc.extend(["<", trp._edge, trp._from])
                else:
                    ret.append(acc)
                    acc = [trp._from, trp._edge, ">", trp._to]
                return jn(trps, acc, ret)
            pass
        if not self._pattern:
            trps = clone(self.triples)
            grps = jn(trps, [], [])
            pats = []
            for grp in grps:
                pats.append(
                    "".join([x.pattern() if not isinstance(x, str) else x
                             for x in grp])
                    )
            self._pattern = ", ".join(pats)
        return self._pattern

    def condition(self):
        return self.pattern()

    def Return(self):
        s = ()
        for t in self.triples:
            s.add(t._from, t._to)
        return ", ".join([_return(x) for x in s if x.var])

# modifiers


def _as(ent, alias):
    """Return copy of ent with As alias set."""
    if isinstance(ent, T):
        return ent
    ret = clone(ent)
    ret.As = alias
    return ret


def _plain(ent):
    """Return entity without properties."""
    ret = None
    if isinstance(ent, (N, R)):
        ret = clone(ent)
        ret.props = {}
    elif isinstance(ent, P):
        ret = clone(ent)
        ret.value = None
    elif isinstance(ent, T):
        ret = clone(ent)
        ret._from.props = {}
        ret._to.props = {}
        ret._edge.props = {}
    else:
        return ent
    return ret


def _anon(ent):
    """Return entity without variable name."""
    ret = clone(ent)
    ret.var = ""
    return ret


def _var(ent):
    """Return entity without label or type."""
    ret = None
    if isinstance(ent, (N, R)):
        ret = clone(ent)
        if hasattr(ret, "label"):
            ret.label = None
        if hasattr(ret, "Type"):
            ret.Type = None
    elif isinstance(ent, T):
        ret = clone(ent)
        ret._from.label = None
        ret._to.label = None
        # leave the edge type alone in a triple...
    else:
        return ent
    return ret

def _plain_var(ent):
    """Return entity with only the variable, no label or properties"""
    return _plain(_var(ent))

# rendering contexts


def _pattern(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.pattern()


def _condition(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.condition()


def _return(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.Return()



"""
mdb.writeable: subclass of `class:bento_meta.MDB` to support writing to an MDB
"""
from functools import wraps
from bento_meta.mdb import make_nanoid
from bento_meta.mdb import MDB

def write_txn(func):
    """Decorates a query function to run a write transaction based
    on its query.
    Query function should return a tuple (qry_string, param_dict)."""
    @wraps(func)
    def wr(self, *args, **kwargs):
        def txn_q(tx):
            (qry,parms)=func(self, *args, **kwargs)
            result = tx.run(qry,parameters=parms)
            return [rec for rec in result]
        with self.driver.session() as session:
            result = session.write_transaction(txn_q)
            return result
    return wr

class WriteableMDB(MDB):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @write_txn
    def put_with_statement(self, qry, parms={}):
        """Run an arbitrary write statement."""
        if not isinstance(qry, str):
            raise RuntimeError("qry= must be a string")
        if not isinstance(parms, dict):
            raise RuntimeError("parms= must be a dict")
        return (qry, parms)


    @write_txn
    def put_term_with_origin(self, term, commit="", _from=1):
        """Merge a bento-meta Term object, that has an Origin object set,
        into an MDB. If a new term is created, assign a random 6-char nanoid 
        to it. The Origin must already be represented in the database.
        :param Term term: Term object
        :param str commit: GitHub commit SHA1 associated with the term (if any)"""
        qry = (
            "match (o:origin {name:$o_name, nanoid:$o_id}) "
            "where not exists(o._to) "
            "merge (t:term {"
            "  value:$t_value,"
            "  origin_id:$t_oid,"
            "  origin_version:$t_oversion,"
            "  origin_definition:$t_odefn,"
            "  _from:$from, _commit:$commit}) "
            "merge (o)<-[:has_origin]-(t) "
            "on create set t.nanoid = $t_id "
            "return t.nanoid"
            )
        parms = {"o_name": term.origin.name, "o_id": term.origin.nanoid,
                 "t_value": term.value, "t_oid": term.origin_id,
                 "t_oversion": term.origin_version,
                 "t_odefn": term.origin_definition,
                 "t_id": make_nanoid(), "commit": commit,
                 "from":_from}
        return (qry, parms)



"""
bento_meta.objects
==================

This module contains the subclasses of :class:`Entity` which are used 
in representing the models contained in the `MDB <https://github.com/CBIIT/bento-mdf>`_.

"""
import sys
sys.path.append("..")
from copy import deepcopy
from bento_meta.entity import Entity
# from pdb import set_trace


def mergespec(clsname, attspec, mapspec):
    """Merge subclass attribute and mapping specification dicts with the
    base class's. Not for human consumption.
    """
    spec = deepcopy(attspec)
    spec.update(Entity.attspec_)
    mo = deepcopy(Entity.mapspec_)
    if "label" in mapspec:
        mo["label"] = mapspec["label"]
    if "key" in mapspec:
        mo["key"] = mapspec["key"]
    if "property" in mapspec:
        mo["property"].update(mapspec["property"])
    if "relationship" in mapspec:
        mo["relationship"].update(mapspec["relationship"])
    mo["relationship"]["_next"]["end_cls"] = {clsname}
    mo["relationship"]["_prev"]["end_cls"] = {clsname}
    return (spec, mo)


class Node(Entity):
    """Subclass that models a data node."""

    attspec_ = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "concept": "object",
        "props": "collection",
    }
    mapspec_ = {
        "label": "node",
        "key": "handle",
        "property": {"handle": "handle", "model": "model", "nanoid": "nanoid"},
        "relationship": {
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "props": {"rel": ":has_property>", "end_cls": "Property"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        },
    }
    (attspec, _mapspec) = mergespec("Node", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)


class Property(Entity):
    """Subclass that models a property of a node or relationship (edge)."""

    attspec_ = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "value_domain": "simple",
        "units": "simple",
        "pattern": "simple",
        "item_domain": "simple",
        "is_required": "simple",
        "concept": "object",
        "value_set": "object",
    }
    mapspec_ = {
        "label": "property",
        "key": "handle",
        "property": {
            "handle": "handle",
            "model": "model",
            "nanoid": "nanoid",
            "value_domain": "value_domain",
            "pattern": "pattern",
            "units": "units",
            "item_domain": "item_domain",
            "is_required": "is_required",
        },
        "relationship": {
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "value_set": {"rel": ":has_value_set>", "end_cls": "ValueSet"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        },
    }
    (attspec, _mapspec) = mergespec("Property", attspec_, mapspec_)
    defaults = { "value_domain":"TBD" }
    
    def __init__(self, init=None):
        super().__init__(init=init)

    @property
    def terms(self):
        """If the `Property` has a ``value_set`` domain, return the `Term` objects
        of its `ValueSet`"""
        if self.value_set:
            return self.value_set.terms
        else:
            return None

    @property
    def values(self):
        """If the `Property` as a ``value_set`` domain, return its term values as a list of str.
        :return: list of term values
        :rtype: list
        """
        if self.value_set:
            return [self.terms[x].value for x in self.terms]


class Edge(Entity):
    """Subclass that models a relationship between model nodes."""
    defaults = {
        "multiplicity": "many_to_many",
    }
    attspec_ = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "multiplicity": "simple",
        "is_required": "simple",
        "src": "object",
        "dst": "object",
        "concept": "object",
        "props": "collection",
    }
    mapspec_ = {
        "label": "relationship",
        "key": "handle",
        "property": {
            "handle": "handle",
            "model": "model",
            "nanoid": "nanoid",
            "multiplicity": "multiplicity",
            "is_required": "is_required",
        },
        "relationship": {
            "src": {"rel": ":has_src>", "end_cls": "Node"},
            "dst": {"rel": ":has_dst>", "end_cls": "Node"},
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "props": {"rel": ":has_property>", "end_cls": "Property"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        },
    }
    (attspec, _mapspec) = mergespec("Edge", attspec_, mapspec_)
     
    def __init__(self, init=None):
        super().__init__(init=init)

    @property
    def triplet(self):
        """A 3-tuple that fully qualifies the edge: ``(edge.handle, src.handle, dst.handle)``
        ``src`` and ``dst`` attributes must be set.
        """
        if self.handle and self.src and self.dst:
            return (self.handle, self.src.handle, self.dst.handle)


class Term(Entity):
    """Subclass that models a term from a terminology."""

    attspec_ = {
        "handle": "simple",
        "value": "simple",
        "nanoid": "simple",
        "origin_id": "simple",
        "origin_version": "simple",
        "origin_definition": "simple",
        "origin_name": "simple",
        "concept": "object",
        "origin": "object",
    }
    mapspec_ = {
        "label": "term",
        # note using 'value' as the key in terms collections may break silently
        # if the terms are coming from different origins, but use the same
        # value (string representation in data) - i.e., value is not necessarily
        # unique - should be the nanoid of the term.
        "key": "value",
        "property": {
            "handle": "handle",
            "value": "value",
            "nanoid": "nanoid",
            "origin_id": "origin_id",
            "origin_version": "origin_version",
            "origin_definition": "origin_definition",
            "origin_name": "origin_name",
        },
        "relationship": {
            "concept": {"rel": ":represents>", "end_cls": "Concept"},
            "origin": {"rel": ":has_origin>", "end_cls": "Origin"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        },
    }
    (attspec, _mapspec) = mergespec("Term", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)


# for ValueSet - updating terms prop should dirty the connected Property
# (from Bento::Meta), signal need to refresh. Engineer so this happens
# here (__setattr__ override), not in Entity
class ValueSet(Entity):
    """Subclass that models an enumerated set of :class:`Property` values.
    Essentially a container for :class:`Term` instances.
    """
    attspec_ = {
        "handle": "simple",
        "nanoid": "simple",
        "url": "simple",
        "prop": "object",
        "origin": "object",
        "terms": "collection",
    }
    mapspec_ = {
        "label": "value_set",
        "property": {"handle": "handle", "url": "url", "nanoid": "nanoid",},
        "relationship": {
            "prop": {"rel": "<:has_value_set", "end_cls": "Property"},
            "terms": {"rel": ":has_term>", "end_cls": "Term"},
            "origin": {"rel": ":has_origin>", "end_cls": "Origin"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        },
    }
    (attspec, _mapspec) = mergespec("ValueSet", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)

    # @property
    # def dirty(self):
    #   return self.pvt.dirty
    # @dirty.setter
    # def dirty(self, value):
    #   print("Dirty, dirty value set!")
    #   self.pvt.dirty = value
    #   if self.prop and value != 0:
    #     self.prop.dirty=1
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name == "dirty":
            if self.prop and value != 0:
                self.prop.dirty = 1


class Concept(Entity):
    """Subclass that models a semantic concept."""

    attspec_ = {"terms": "collection"}
    mapspec_ = {
        "label": "concept",
        "relationship": {
            "terms": {"rel": "<:represents", "end_cls": "Term"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}            
        },
    }
    (attspec, _mapspec) = mergespec("Concept", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)

class Predicate(Entity):
    """Subclass that models a semantic link between concepts."""
    attspec_ = {
        "handle": "simple",
        "subject": "object",
        "object": "object"}
    mapspec_ = {
        "label": "predicate",
        "key": "handle",
        "property": {
            "handle": "handle",
        },
        "relationship": {
            "subject": {"rel": ":has_subject>", "end_cls": "Concept"},
            "object": {"rel": ":has_object>", "end_cls": "Concept"},
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}            
        },
    }
    (attspec, _mapspec) = mergespec("Predicate", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)
        

class Origin(Entity):
    """Subclass that models a :class:`Term` 's authoritative source."""

    attspec_ = {"url": "simple", "is_external": "simple", "name": "simple", "nanoid": "simple",}
    mapspec_ = {
        "label": "origin",
        "key": "name",
        "property": {
            "name": "name",
            "url": "url",
            "is_external": "is_external",
            "nanoid": "nanoid",
        },
        "relationship": {
            "tags": {"rel": ":has_tag>", "end_cls":"Tag"}
        }
    }
    (attspec, _mapspec) = mergespec("Origin", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)


class Tag(Entity):
    """Subclass that allows simple key-value tagging of a model at arbitrary points."""

    attspec_ = {"key":"simple", "value": "simple"}
    mapspec_ = {
        "label": "tag",
        "key": "key",
        "property": {"key": "key", "value": "value"},
    }
    (attspec, _mapspec) = mergespec("Tag", attspec_, mapspec_)

    def __init__(self, init=None):
        super().__init__(init=init)



"""
bento_meta.model
================

This module contains :class:`Model`, a class for managing data models housed
in the Bento Metamodel Database. Models are built from `bento_meta.Entity`
subclasses (see :mod:`bento_meta.objects`). A Model can be used with or 
without a Neo4j database connection.

"""
import re
import sys
sys.path.append("..")
from uuid import uuid4
from warnings import warn
import neo4j.graph
from bento_meta.mdb import MDB, make_nanoid
from bento_meta.object_map import ObjectMap
from bento_meta.entity import Entity, ArgError
from bento_meta.objects import (
    Node,
    Property,
    Edge,
    Term,
    ValueSet,
    Concept,
    Predicate,
    Origin,
    Tag,
)
# from pdb import set_trace


class Model(object):
    def __init__(self, handle=None, mdb=None):
        """Model constructor.

        :param str handle: A string name for the model. Corresponds to the model property in MDB database nodes.
        :param bento_meta.mdb.MDB mdb: An MDB object containing the db connection (see :class:`bento_meta.mdb.MDB`)
        """
        if not handle:
            raise ArgError("model requires arg 'handle' set")
        self.handle = handle
        self._mdb = None
        self.nodes = {}
        self.edges = {}  # keys are (edge.handle, src.handle, dst.handle) tuples
        self.props = {}  # keys are ({edge|node}.handle, prop.handle) tuples
        self.terms = {}
        self.removed_entities = []

        if mdb:
            self.mdb = mdb

    @classmethod
    def versioning(cls, on=None):
        """Get or set versioning state.

        :param boolean on: True, apply versioning. False, do not.

        Note: this delegates to :meth:`Entity.versioning`.
        """
        if on is None:
            return Entity.versioning_on
        Entity.versioning_on = on
        return Entity.versioning_on

    @classmethod
    def set_version_count(cls, ct):
        """Set the integer version counter.

        :param int ct: Set version counter to this value.

        Note: this delegates to :meth:`Entity.set_version_count`.
        """
        Entity.set_version_count(ct)

    @property
    def drv(self):
        """Neo4j database driver from MDB object"""
        return self._mdb.driver if self._mdb else None

    @property
    def mdb(self):
        return self._mdb
    
    @mdb.setter
    def mdb(self, value):
        if isinstance(value, MDB):
            self._mdb = value
            for cls in (Node, Property, Edge, Term, ValueSet, Concept, Predicate, Origin, Tag):
                cls.object_map = ObjectMap(cls=cls, drv=value.driver)
        elif not value:
            self._mdb = None
            for cls in (Node, Property, Edge, Term, ValueSet, Concept, Origin, Tag):
                cls.object_map = None
        else:
            raise ArgError(
                "mdb= arg must be a bento_meta.mdb.MDB object"
            )

    def add_node(self, node=None):
        """Add a :class:`Node` to the model.

        :param Node node: A :class:`Node` instance, a :class:`neo4j.graph.Node`, or a dict

        The model attribute of ``node`` is set to `Model.handle`
        """
        if not node:
            raise ArgError("arg must be Node, dict, or graph.Node")
        if isinstance(node, (dict, neo4j.graph.Node)):
            node = Node(node)
        if not node.model:
            node.model = self.handle
        for p in node.props.values():
            self.add_prop(node, p)
        self.nodes[node.handle] = node
        return node

    def add_edge(self, edge=None):
        """Add an :class:`Edge` to the model.

        :param Edge edge: A :class:`Edge` instance, a :class:`neo4j.graph.Node`, or a dict

        The model attribute of ``edge`` is set to `Model.handle`
        """
        if not edge:
            raise ArgError("arg must be Edge, dict, or graph.Node")
        if isinstance(edge, (dict, neo4j.graph.Node)):
            edge = Edge(edge)
        if not edge.src or not edge.dst:
            raise ArgError("edge must have both src and dst set")
        if not edge.model:
            edge.model = self.handle
        if not self.contains(edge.src):
            warn("Edge source node not yet in model; adding it")
            self.add_node(edge.src)
        if not self.contains(edge.dst):
            warn("Edge destination node not yet in model; adding it")
            self.add_node(edge.dst)
        for p in edge.props.values():
            self.add_prop(edge, p)
        self.edges[edge.triplet] = edge
        return edge

    def add_prop(self, ent, prop=None):
        """Add a :class:`Property` to the model.

        :param Node|Edge ent: Attach ``prop`` to this entity
        :param Property prop: A :class:`Property` instance, a :class: `neo4j.graph.Node`, or a dict
        :param boolean reuse: If True, reuse existing property with same handle

        The model attribute of ``prop`` is set to `Model.handle`. Within a model,
        :class:`Property` entities are unique with respect to their
        handle (but can be reused). This method will look for an existing
        property within the model with the given handle, and add an item to
        Model.props pointing to it if found.
        """
        if not isinstance(ent, (Node, Edge)):
            raise ArgError("arg 1 must be Node or Edge")
        if not prop:
            raise ArgError("arg 2 must be Property, dict, or graph.Node")
        if isinstance(prop, (dict, neo4j.graph.Node)):
            prop = Property(prop)
        if not prop.model:
            prop.model = self.handle
        key = [ent.handle] if isinstance(ent, Node) else list(ent.triplet)
        key.append(prop.handle)
        ent.props[getattr(prop, type(prop).mapspec()["key"])] = prop
        if not tuple(key) in self.props:
            self.props[tuple(key)] = prop
        return prop

    def annotate(self, ent, term):
        """
        Associate a single :class:`Term` with an :class:`Entity`. This creates a Concept entity
        if needed and links both the Entity and the Term to the concept, in keeping with the MDB
        spec. It supports the Term key in MDF.
        :param Entity ent: :class:`Entity` object to annotate
        :param Term term: :class:`Term` object to describe the Entity
        """
        if not isinstance(ent, Entity):
            raise ArgError("arg1 must be Entity")
        if not isinstance(term, Term):
            raise ArgError("arg2 must be Term")
        if not ent.concept:
            ent.concept = Concept({"nanoid":make_nanoid()})
        if (term.value, term.origin_name) in ent.concept.terms:
            raise ValueError("Concept already represented by a Term with value '{}'"
                             "and origin_name '{}'".format(term.value, term.origin_name))
        ent.concept.terms[(term.value, term.origin_name)] = term
        
    def add_terms(self, prop, *terms):
        """Add a list of :class:`Term` and/or strings to a :class:`Property` with a value domain of ``value_set``

        :param Property prop: :class:`Property` to modify
        :param list terms: A list of :class:`Term` instances and/or str

        :class:`Term` instances are created for strings; `Term.value` is set to the string.
        """
        if not isinstance(prop, Property):
            raise ArgError("arg1 must be Property")
        if not re.match("value_set|enum", prop.value_domain):
            raise AttributeError(
                "Property value domain is not value_set or enum, can't add terms"
            )
        if not prop.value_set:
            warn("Creating ValueSet object for Property " + prop.handle)
            prop.value_set = ValueSet({"prop": prop, "_id": str(uuid4())})
            prop.value_set.handle = self.handle + prop.value_set._id[0:8]

        for t in terms:
            if isinstance(t, str):
                warn("Creating Term object for string '{term}'".format(term=t))
                t = Term({"value": t})
            elif not isinstance(t, Term):
                raise ArgError("encountered arg that was not a str or Term object")
            prop.value_set.terms[t.value] = t

    def rm_node(self, node):
        """Remove a :class:`Node` from the Model instance.

        :param Node node: Node to be removed

        Note: A node can't be removed if it is participating in an edge (i.e.,
        if the node is some edge's src or dst attribute)

        *Clarify what happens in the Model object, in the database when versioning
        is off, in the database when versioning is on*
        """
        if not isinstance(node, Node):
            raise ArgError("arg must be a Node object")
        if not self.contains(node):
            warn(
                "node '{node}' not contained in model '{model}'".format(
                    node=node.handle, model=self.handle
                )
            )
            return
        if self.edges_by_src(node) or self.edges_by_dst(node):
            raise ValueError(
                "can't remove node '{node}', it is participating in edges".format(
                    node=node.handle
                )
            )
        for p in node.props:
            try:
                del self.props[(node.handle, p.handle)]
            except:
                pass
        del self.nodes[node.handle]
        self.removed_entities.append(node)
        return node

    def rm_edge(self, edge):
        """Remove an :class:`Edge` instance from the Model instance.

        :param Edge edge: Edge to be removed

        *Clarify what happens in the Model object, in the database when versioning
        is off, in the database when versioning is on*
        """
        if not isinstance(edge, Edge):
            raise ArgError("arg must be an Edge object")
        if not self.contains(edge):
            warn(
                "edge '{edge}' not contained in model '{model}'".format(
                    edge=edge.triplet, model=self.handle
                )
            )
            return
        for p in edge.props:
            try:
                k = list(edge.triplet)
                k.append(p.handle)
                del self.props[tuple(k)]
            except:
                pass
        del self.edges[edge.triplet]
        edge.src = None
        edge.dst = None
        self.removed_entities.append(edge)
        return edge

    def rm_prop(self, prop):
        """Remove a :class:`Property` instance from the Model instance.

        :param Property prop: Property to be removed

        *Clarify what happens in the Model object, in the database when versioning
        is off, in the database when versioning is on*
        """
        if not isinstance(prop, Property):
            raise ArgError("arg must be a Property object")
        if not self.contains(prop):
            warn(
                "prop '{prop}' not contained in model '{model}'".format(
                    prop=prop.handle, model=self.handle
                )
            )
            return
        for okey in prop.belongs:
            owner = prop.belongs[okey]
            (i, att, key) = okey
            getattr(owner, att)[key] == None
            k = [owner.handle] if isinstance(owner, Node) else list(owner.triplet)
            k.append(key)
            del self.props[tuple(k)]
        self.removed_entities.append(prop)
        pass

    def rm_term(self, term):
        """Not implemented."""
        if not isinstance(term, Term):
            raise ArgError("arg must be a Term object")
        pass

    def assign_edge_end(self, edge=None, end=None, node=None):
        """Move the src or dst of an :class:`Edge` to a different :class:`Node`.

        :param Edge edge: Edge to manipulate
        :param str end: Edge end to change (src|dst)
        :param Node node: Node to be connected

        Note: Both ``node`` and ``edge`` must be present in the Model instance
        (via :meth:`add_node` and :meth:`add_edge`)
        """
        if not isinstance(edge, Edge):
            raise ArgError("edge= must an Edge object")
        if not isinstance(node, Node):
            raise ArgError("node= must a Node object")
        if end not in ["src", "dst"]:
            raise ArgError("end= must be one of 'src' or 'dst'")
        if not self.contains(edge) or not self.contains(node):
            warn("model must contain both edge and node")
            return
        del self.edges[edge.triplet]
        setattr(edge, end, node)
        self.edges[edge.triplet] = edge
        return edge

    def contains(self, ent):
        """Ask whether an entity is present in the Model instance.

        :param Entity ent: Entity in question

        Note: Only works on Nodes, Edges, and Properties
        """
        if not isinstance(ent, Entity):
            warn("argument is not an Entity subclass")
            return
        if isinstance(ent, Node):
            return ent in set(self.nodes.values())
        if isinstance(ent, Edge):
            return ent in set(self.edges.values())
        if isinstance(ent, Property):
            return ent in set(self.props.values())
        pass

    def edges_in(self, node):
        """Get all :class:`Edge` that have a given :class:`Node` as their dst attribute

        :param Node node: The node
        :return: list of :class:`Edge`
        """
        if not isinstance(node, Node):
            raise ArgError("arg must be Node")
        return [self.edges[i] for i in self.edges if i[2] == node.handle]
        pass

    def edges_out(self, node):
        """Get all :class:`Edge` that have a given :class:`Node` as their src attribute

        :param Node node: The node
        :return: list of :class:`Edge`
        """
        if not isinstance(node, Node):
            raise ArgError("arg must be Node")
        return [self.edges[i] for i in self.edges if i[1] == node.handle]
        pass

    def edges_by(self, key, item):
        if key not in ["src", "dst", "type"]:
            raise ArgError("arg 'key' must be one of src|dst|type")
        if isinstance(item, Node):
            idx = 1 if key == "src" else 2
            return [self.edges[x] for x in self.edges if x[idx] == item.handle]
        else:
            return [self.edges[x] for x in self.edges if x[0] == item]

    def edges_by_src(self, node):
        """Get all :class:`Edge` that have a given :class:`Node` as their src attribute

        :param Node node: The node
        :return: list of :class:`Edge`
        """
        return self.edges_by("src", node)

    def edges_by_dst(self, node):
        """Get all :class:`Edge` that have a given :class:`Node` as their dst attribute

        :param Node node: The node
        :return: list of :class:`Edge`
        """
        return self.edges_by("dst", node)

    def edges_by_type(self, edge_handle):
        """Get all :class:`Edge` that have a given edge type (i.e., handle)

        :param str edge_handle: The edge type
        :return: list of :class:`Edge`
        """
        if not isinstance(edge_handle, str):
            raise ArgError("arg must be str")
        return self.edges_by("type", edge_handle)

    def dget(self, refresh=False):
        """Pull model from MDB into this Model instance, based on its handle

        Note: is a noop if `Model.mdb` is unset.
        """
        if not self.mdb:
            return
        if refresh:
            ObjectMap.clear_cache()
        with self.drv.session() as session:
            result = session.run(
                "match p = (s:node)<-[:has_src]-(r:relationship)-[:has_dst]->(d:node) "
                "where s.model=$hndl and r.model=$hndl and d.model=$hndl return p",
                {"hndl": self.handle},
            )
            for rec in result:
                (ns, nr, nd) = rec["p"].nodes
                ns = Node(ns)
                nr = Edge(nr)
                nd = Node(nd)
                ObjectMap.cache[ns.neoid] = ns
                ObjectMap.cache[nr.neoid] = nr
                ObjectMap.cache[nd.neoid] = nd
                nr.src = ns
                nr.dst = nd
                self.nodes[ns.handle] = ns
                self.nodes[nd.handle] = nd
                self.edges[nr.triplet] = nr

        with self.drv.session() as session:
            result = session.run(
                "match (n:node)-[:has_property]->(p:property) where n.model=$hndl and p.model=$hndl return id(n), p",
                {"hndl": self.handle},
            )
            for rec in result:
                n = ObjectMap.cache.get(rec["id(n)"])
                if n is None:
                    warn(
                        "node with id {nid} not yet retrieved".format(nid=rec["id(n)"])
                    )
                    continue
                p = Property(rec["p"])
                ObjectMap.cache[p.neoid] = p
                self.props[(n.handle, p.handle)] = p
                n.props[p.handle] = p
                p.dirty = -1
        with self.drv.session() as session:
            result = session.run(
                "match (r:relationship)-[:has_property]->(p:property) "
                "where r.model=$hndl and p.model=$hndl return id(r), p",
                {"hndl": self.handle},
            )
            for rec in result:
                e = ObjectMap.cache.get(rec["id(r)"])
                if e is None:
                    warn(
                        "relationship with id {rid} not yet retrieved".format(
                            rid=rec["id(r)"]
                        )
                    )
                    continue
                p = Property(rec["p"])
                ObjectMap.cache[p.neoid] = p
                k = list(e.triplet)
                k.append(p.handle)
                self.props[tuple(k)] = p
                e.props[p.handle] = p
                p.dirty = -1
        return self

    def dput(self):
        """Push this Model's objects to MDB.

        Note: is a noop if `Model.mdb` is unset.
        """
        if not self.mdb:
            return
        seen = {}

        def do_(obj):
            if id(obj) in seen:
                return
            seen[id(obj)] = 1
            if obj.dirty == 1:
                obj.dput()
            atts = [x for x in type(obj).attspec if type(obj).attspec[x] == "object"]
            for att in atts:
                ent = getattr(obj, att)
                if ent:
                    do_(ent)
            atts = [
                x for x in type(obj).attspec if type(obj).attspec[x] == "collection"
            ]
            for att in atts:
                ents = getattr(obj, att)
                if ents:
                    for ent in ents:
                        do_(ents[ent])

        for e in self.removed_entities:
            # detach
            with self.drv.session() as session:
                result = session.run(
                    "match (e)-[r]-() where id(e)=$eid delete r return id(e)",
                    {"eid": e.neoid},
                )
                for rec in result:
                    pass
        for e in self.nodes.values():
            do_(e)
        for e in self.edges.values():
            do_(e)
        for e in self.props.values():
            do_(e)
        return



import re
from bento_meta.util.cypher.entities import (  # noqa E402
    N, R, P, N0, R0, G,
    _as, _var, _plain, _anon,
    )

from bento_meta.util.cypher.functions import (
    count, exists, labels, group, And, Or, Not,
    )
from bento_meta.util.cypher.clauses import (
    Match, Where, With, Return,
    Statement
    )
from pdb import set_trace  # noqa E402

avail_funcs = {x.__name__: x for x in (count, exists, labels, group, And, Or, Not)}


class _engine(object):
    paths = None

    def __init__(self, use_params=True):
        self.use_params = use_params
        self.error = None
        self.statement = None
        self.params = None
        self.key = ""
        self.path_id = None
        pass

    @classmethod
    def set_paths(cls, paths):
        cls.paths = paths

    def parse(self, toks):
        toks = toks
        pth = self.paths
        return self._walk(None, toks, pth)

    def _process_node(self, block):
        ret = None
        if isinstance(block, str):
            if block == '_var':
                ret = N()
            else:
                ret = N(label=block)
        elif isinstance(block, dict):
            if not block['_label']:
                self.error = {
                    "description": "_node block requires _label key",
                    "block": block,
                    }
                return False
            ret = N(label=block['_label'])
            if block.get('_prop'):
                ret._add_props(self._process_prop(block['_prop']))
            if block.get('_edge') and block.get('_node'):
                n = self._process_node(block['_node'])
                e = self._process_edge(block['_edge'])
                ret = G(ret, e, n)
        else:
            self.error = {
                "description": "Can't process _node block",
                "block": block,
                }
            return False
        return ret

    def _process_prop(self, block, value=None):
        ret = None
        if isinstance(block, str):
            ret = P(handle=block,
                    value=value)
        elif (isinstance(block, dict) and
              block.get('_handle') and block.get('_value')):
            ret = P(handle=block['_handle'],
                    value=block['_value'])
        else:
            self.error = {
                "description": "Can't process _prop block",
                "block": block
                }
            return False
        return ret

    def _process_edge(self, block):
        ret = None
        if isinstance(block, str):
            ret = R(Type=block)
        elif isinstance(block, dict):
            ret = R(Type=block['_type'])
            if block.get('_dir'):
                ret._dir = block.get('_dir')
            if block.get('_join'):
                # when an edge connects two complex paths
                # define the nodes on the incoming path and new path
                # to link with this edge
                ret._join = block['_join']
        else:
            self.error = {
                "description": "Can't process _edge block",
                "block": block,
                }
            return False
        return ret

    def _process_func(self, block):
        ret = {}
        if isinstance(block, str):
            as_ = re.split("@",block)
            ret['_func'] = as_[0]
            if len(as_) > 1:
                as_ = as_[1]
            else:
                as_ = None
            if ret['_func'] in avail_funcs:
                # the Func subclass:
                ret['_func'] = avail_funcs[ret['_func']]
                ret['_func_as'] = as_
            else:
                self.error = {
                    "description": "Sorry, no cypher function '{}' is currently defined".format(ret['_func']),
                    "block": block,
                    }
                return False  # ERR no such function available
        else:
            self.error = {
                "description": "Block type other than simple string not yet handled for _func",
                "block": block,
                }
            return False
        return ret
                
    def _create_statement(self, ent, pad):
        match_clause = Match(ent)
        ret_clause = None
        a = []
        if isinstance(pad['_return'], str):
            if pad['_return'] == '_items':
                if isinstance(ent, (N, R)):
                    ret_clause = Return(ent)
                else:
                    ret_clause = Return(*ent.nodes())
            else:
                if isinstance(ent, N) and ent.label == pad['_return'] and ent.var:
                    a.append(ent)
                elif isinstance(ent, R) and ent.Type == pad['_return'] and ent.var:
                    a.append(ent)
                else:
                    a.extend([x for x in ent.nodes()
                              if x.label == pad['_return'] and x.var])
                if not a:
                    self.error = {
                        "description": "No named node to return with label '{}'".format(pad['_return']),
                        "ent": ent,
                        "pad": pad,
                        }
                    return False
                else:
                    ret_clause = Return(*a)
        elif isinstance(pad['_return'], dict):
            retblock = pad['_return']
            if retblock.get('_path_id'):
                self.path_id = retblock['_path_id']
            if retblock.get('_nodes'):
                # assume is list-valued
                if not isinstance(retblock['_nodes'], list):
                    self.error = {
                        "description": '_nodes key must point to a list',
                        "ent": ent,
                        "pad": pad,
                        }
                    return False
                labels = {x[0]:x[1] for x in [re.split("@",y)+[None] for y in retblock['_nodes']]}
                if '*' in labels:
                    a.append('*')
                elif (isinstance(ent, N) and ent.var):
                    if ent.label in labels:
                        if labels[ent.label]:
                            a.append(_as(ent, labels[ent.label]))
                        else:
                            a.append(ent)
                    if '_var' in labels and ent == pad['_var']:
                        if labels['_var']:
                            a.append(_as(ent, labels['_var']))
                        else:
                            a.append(ent)
                else:
                    for n in ent.nodes():
                        if n.label in labels and n.var:
                            if labels[n.label]:
                                a.append(_as(n, labels[n.label]))
                            else:
                                a.append(n)
                        if '_var' in labels and n == pad['_var']:
                            if labels['_var']:
                                a.append(_as(n, labels['_var']))
                            else:
                                a.append(n)
                                        
            if retblock.get('_edges'):
                if not isinstance(retblock['_edges'], list):
                    self.error = {
                        "description": '_edges key must point to a list',
                        "ent": ent,
                        "pad": pad,
                        }
                    return False
                types = {x[0]:x[1] for x in [re.split("@",y)+[None] for y in retblock['_edges']]}
                if isinstance(ent, R) and ent.var:
                    if ent.Type in types:
                        if types[ent.Type]:
                            _as(ent, types[ent.Type])
                        else:
                            a.append(ent)
                    if '_var' in types and ent == pad['_var']:
                        if types['_var']:
                            a.append(_as(ent, types['_var']))
                        else:
                            a.append(ent)
                else:
                    for e in ent.edges():
                        if e.Type in types and e.var:
                            if types[e.Type]:
                                a.append(_as(e, types[e.Type]))
                            else:
                                a.append(e)
            if not a:
                self.error = {
                    "description": "No named nodes or edges matching the path _return spec",
                    "ent": ent,
                    "pad": pad,
                    }
                return False
            else:
                if retblock.get('_func'):
                    f = self._process_func(retblock['_func'])
                    if not f:
                        return False  # bad _func spec
                    a = [f['_func'](x) for x in a]
                    if f.get('_func_as'):
                        a = [_as(x, f['_func_as']) for x in a]
                ret_clause = Return(*a)
        else:
            self.error = {
                "description": "_return specification not str or dict",
                "ent": ent,
                "pad": pad,
                }
            return False
        self.statement = Statement(match_clause, ret_clause,
                                   use_params=self.use_params)
        self.params = self.statement.params
        return True

    def _walk(self, ent, toks, pth):
        if not toks or not pth:
            self.error = {
                "description": "_walk: Either toks or pth is empty"
                }
            return False
        tok = toks[0]
        pad = {}
        parm = None
        if tok in pth:
            # plain token - shouldn't start with _
            self.key = "/".join([self.key, tok]) if self.key else tok
            pth = pth[tok]
        elif any([x.startswith('$') for x in pth]):
            # parameter
            parm = [x for x in pth if x.startswith('$')][0]
            pth = pth[parm]
            # load pad
            pad['_prop'] = P(handle=parm[1:], value=tok)
            # note that the cache key will be interpreted as a regexp
            self.key = "/".join([self.key, "([a-zA-Z0-9_]+)"]) if self.key else "([a-zA-Z0-9_]+)"
            pass
        else:
            self.error = {
                "description": "Token '{}' not on valid path".format(tok),
                "token": tok
                }
            return False
        # collect/create items req by block in the pad, and then
        # execute operations on these items in standard order below.
        for opn in [x for x in pth if x.startswith('_')]:
            # operations in block
            if opn == "_node":
                if parm:  # processing a parameter
                    pad['_node'] = pth['_node']
                else:
                    pad['_node'] = self._process_node(pth['_node'])
                    if pth['_node'] == '_var':
                        pad['_var'] = pad['_node']
            elif opn == "_edge":
                if parm:  # processing a parameter
                    pad['_edge'] = pth['_edge']
                else:
                    pad['_edge'] = self._process_edge(pth['_edge'])
                    if pth['_edge'] == '_var':
                        pad['_var'] == pad['_edge']
            elif opn == "_prop":
                # WARN or ERR if pad.get('_prop') is True
                # - a parm was already handled in that case
                pad['_prop'] = self._process_prop(pth['_prop'], value=tok)
            elif opn == "_return":
                if len(toks) == 1:  # we're on the last token
                    # pad['_return'] not empty means we're finished
                    pad['_return'] = pth['_return']
                else:  # more toks to go...
                    pass  # so ignore it
                pass
            elif opn == "_func":
                _func = self._process_func(pth['_func'])
                if not _func:
                    return False # bad _func spec
                for k in _func:
                    pad[k] = _func[k]
            pass

        # pad ready for operations
        new_ent = None
        if (pad.get('_prop')):  # add props to incoming entity
            if isinstance(ent, (N, R)):
                ent._add_props(pad['_prop'])
            else:
                if not pad.get('_node') and not pad.get('_edge'):
                    self.error = {
                        "description": "Both _edge and _node must be defined here",
                        "token": tok
                        }
                    return False  # ERR, if incoming is path, need to specify the node which gets the property (by label)
                else:
                    if pad.get('_node'):
                        n = [x for x in ent.nodes()
                             if x.label == pad['_node']]
                        if not n:
                            self.error = {
                                "description": "Node specified by _node is not present",
                                "token": tok
                                }
                            return False  # specified node can't be found in ent
                        else:
                            n[0]._add_props(pad['_prop'])
                            pad['_node'] = None
                            pad['_prop'] = None
                    if pad.get('_edge'):
                        e = [x for x in ent.edges()
                             if x.Type == pad['_edge']]
                        if not e:
                            self.error = {
                                "description": "Edge specified by _edge is not present",
                                "token": tok
                                }
                            return False  # specified edge not found in ent
                        else:
                            e[0]._add_props(pad['_prop'])
                            pad['_edge'] = None
                            pad['_prop'] = None

        if (pad.get('_node')):  # new entity
            new_ent = pad['_node']
        if (pad.get('_edge')):
            if not ent:  # no input entity
                self.error = {
                    "description": "No incoming entity to apply _edge to here",
                    "token": tok
                    }
                return False  # ERR
            if not new_ent:  # no new entity
                self.error = {
                    "description": "No new entity to link to",
                    "token": tok
                    }
                return False  # different ERR
            if isinstance(ent, N) and isinstance(new_ent, N):
                # simple relationship
                new_ent = pad['_edge'].relate(ent, new_ent)
            else:
                new_ent = G(ent, pad['_edge'], new_ent)
        if pad.get('_return'):  # we made it
            if self._create_statement(new_ent or ent, pad):
                return True
            else:
                return False
        else:
            if len(toks) == 1:
                self.error = {
                    "description": "Reached end of path, but found no _return spec",
                    "token": tok
                    }
                return False  # ERR reached end of toks, no _return found
            else:
                return self._walk(new_ent or ent, toks[1:], pth)



"""
bento_meta.entity
=================

This module contains 
* `Entity`, the base class for metamodel objects, 
* the `CollValue` class to manage collection-valued attributes, and 
* the `ArgError` exception.
"""
# from pdb import set_trace
from warnings import warn
from collections import UserDict


class ArgError(Exception):
    """Exception for method argument errors"""
    pass


class Entity(object):
    """Base class for all metamodel objects.

    Entity contains all the magic for metamodel objects such as
    `bento_meta.objects.Node` and 'bento_meta.object.Edge`. It will rarely
    be used directly. Entity redefines `__setattr__` and `__getattr__` to
    enable necessary bookkeeping for model versioning and graph database
    object mapping under the hood.

    The Entity class also defines private and declared attributes that are
    common to all metamodel objects. It provides the machinery to manage
    private attributes separately from declared attributes, and to raise
    exceptions when attempts are made to access attributes that are not
    declared.
    """
    pvt_attr = [
        "pvt",
        "neoid",
        "dirty",
        "removed_entities",
        "attspec",
        "mapspec",
        "belongs",
    ]
    defaults = {},
    attspec_ = {
        "_id": "simple",
        "nanoid": "simple",
        "desc": "simple",
        "_next": "object",
        "_prev": "object",
        "_from": "simple",
        "_to": "simple",
        "_commit": "simple",
        "tags": "collection",
    }
    attspec = attspec_
    mapspec_ = {
        "label": None,
        "key": "_id",
        "property": {"_id": "id", "desc": "desc",
                     "_from": "_from", "_to": "_to",
                     "_commit": "_commit", "nanoid": "nanoid"},
        "relationship": {
            "_next": {"rel": ":_next>", "end_cls": set()},
            "_prev": {"rel": ":_prev>", "end_cls": set()},
            "tags": {"rel": ":has_tag", "end_cls": {"Tag"}},
        },
    }
    object_map = None
    version_count = None
    versioning_on = False

    def __init__(self, init=None):
        """Entity constructor. Always called by subclasses.
        .. py:function:: Node(init)
        :param dict init: A dict of attribute names and values. Undeclared attributes are ignored.
        :param neo4j.graph.Node init: a `neo4j.graph.Node` object to be stored as a model object.
        :param `bento_meta.Entity` init: an Entity (of matching subclass). Used to duplicate another model object.
        """
        if not set(type(self).attspec.values()) <= set(
            ["simple", "object", "collection"]
        ):
            raise ArgError("unknown attribute type in attspec")

        # private
        self.pvt = {}
        self.neoid = None
        self.dirty = 1
        self.removed_entities = []
        self.belongs = {}
        # merge to universal map - no, do in the subclasses
        # type(self).mergespec()

        if init:
            if isinstance(init, Entity):
                self.set_with_entity(init)
            elif isinstance(init, dict):
                self.set_with_dict(init)
            elif (
                type(init).__name__ == "Node"
            ):  # neo4j.graph.Node - but don't want to import that
                self.set_with_node(init)
        for att in type(self).attspec:
            if att not in self.__dict__:
                if self.attspec[att] == "collection":
                    setattr(self, att, CollValue({}, owner=self, owner_key=att))
                else:
                    setattr(self, att, None)
        if type(self).versioning_on:
            self._from = type(self).version_count

    @classmethod
    def mapspec(cls):
        """The object to database mapping specification. Is a class method, not a property."""
        if not hasattr(cls, "_mapspec"):
            cls.mergespec()
        return cls._mapspec

    @classmethod
    def versioning(cls, on=None):
        """Get or set whether versioning is applied to object manipulations
        :param boolean on: True, apply versioning. False, do not.
        """
        if on is None:
            return cls.versioning_on
        cls.versioning_on = on
        return cls.versioning_on

    @classmethod
    def set_version_count(cls, ct):
        """Set the integer version counter. This will usually be accessed via a
        `bento_meta.Model` instance
        :param int ct: Set version counter to this value
        """
        if not isinstance(ct, int) or ct < 0:
            raise ArgError("arg must be a positive integer")
        cls.version_count = ct

    @classmethod
    def default(cls, propname):
        """Returns a default value for the property named, or None if no default defined."""
        if cls.defaults.get(propname):
            return cls.defaults[propname]
        else:
            return None
        

    # @classmethod
    def get_by_id(self, id):
        """Get an object from the db with the id attribute (not the Neo4j id). Returns a new object.
        :param string id: value of id for desired object
        """
        if self.object_map:
            print('  > now in entity.get_by_id where self is {}'.format(self))
            print('  > and class is {}'.format(self.__class__))
            return self.object_map.get_by_id(self, id)
        else:
            print('    _NO_ cls.object_map detected')
            pass

    @property
    def dirty(self):
        """Flag whether this instance has been changed since retrieval from
        the database
        Set to -1, ensure that the next time an attribute is accessed, the instance
        will retrieve itself from the database.
        """
        return self.pvt["dirty"]

    @dirty.setter
    def dirty(self, value):
        self.pvt["dirty"] = value

    @property
    def versioned(self):
        """Is this instance versioned?"""
        return self._from

    @property
    def removed_entities(self):
        return self.pvt["removed_entities"]

    @property
    def object_map(self):
        return self.pvt["object_map"]

    @property
    def belongs(self):
        """Dict that stores information on the owners (referents) of this instance
        in the model
        """
        return self.pvt["belongs"]

    def clear_removed_entities(self):
        self.pvt["removed_entities"] = []

    def set_with_dict(self, init):
        for att in type(self).attspec:
            if att in init:
                if type(self).attspec[att] == "collection":
                    setattr(self, att, CollValue(init[att], owner=self, owner_key=att))
                else:
                    setattr(self, att, init[att])

    def set_with_node(self, init):
        # this unsets any attribute that is not present in the Node's properties
        for att in [a for a in type(self).attspec if type(self).attspec[a] == "simple"]:
            patt = type(self).mapspec()["property"][att]
            if patt in init:
                setattr(self, att, init[patt])
            else:
                setattr(self, att, None)
        self.neoid = init.id

    def set_with_entity(self, ent):
        if not isinstance(self, type(ent)):
            raise ArgError(
                "class mismatch: I am a {slf}, but arg is a {ent}".format(
                    slf=type(self).__name__, ent=type(ent).__name__
                )
            )
        for k in type(self).attspec:
            atts = type(self).attspec[k]
            if k == "_next" or k == "_prev":
                break
            if atts == "simple":
                setattr(self, k, getattr(ent, k))
            elif atts == "object":
                setattr(self, k, getattr(ent, k))
            elif atts == "collection":
                setattr(self, k, CollValue(getattr(ent, k), owner=self, owner_key=k))
                pass
            else:
                raise RuntimeError("unknown attribute type '{atts}'".format(atts=atts))
        for okey in ent.belongs:
            self.belongs[okey] = ent.belongs[okey]
        self.neoid = ent.neoid
        self.dirty = 1
        return self

    def __getattribute__(self, name):
        if name in type(self).attspec:
            # declared attr, send to __getattr__ for magic
            return self.__getattr__(name)
        else:
            return object.__getattribute__(self, name)

    def __getattr__(self, name):
        if name in Entity.pvt_attr:
            return self.__dict__["pvt"][name]
        elif name in type(self).attspec:
            if name not in self.__dict__ or self.__dict__[name] is None:
                return None
            if type(self).attspec[name] == "object":
                # magic - lazy getting
                if self.__dict__[name].dirty < 0:
                    self.__dict__[name].dget()
            return self.__dict__[name]
        else:
            raise AttributeError(
                "get: attribute '{name}' neither private nor declared for subclass {cls}".format(
                    name=name, cls=type(self).__name__
                )
            )

    def __setattr__(self, name, value):
        if name == "pvt":
            self.__dict__["pvt"] = value
        elif name in Entity.pvt_attr:
            self.__dict__["pvt"][name] = value
        elif name in type(self).attspec:
            self._check_value(name, value)
            if name in ["_prev", "_next", "_from", "_to"]:
                self.dirty = 1
                self.__dict__[name] = value
            else:
                self._set_declared_attr(name, value)
        else:
            raise AttributeError(
                "set: attribute '{name}' neither private nor declared for subclass {cls}".format(
                    name=name, cls=type(self).__name__
                )
            )

    def version_me(setattr_func):
        def _version_set_declared_attr(self, name, value):
            if not type(self).versioning_on:
                return setattr_func(self, name, value)
            if not self.versioned:
                return setattr_func(self, name, value)
            elif (type(self).version_count > self._from) and (self._to is None):
                # dup becomes the "old" object and self the "new":
                dup = self.dup()
                dup._to = type(self).version_count
                dup._from = self._from
                self._from = type(self).version_count
                if self._prev:
                    dup._prev = self._prev
                    self._prev._next = dup
                dup._next = self
                self._prev = dup
                self.neoid = None
                # make the owners own dup, rather than self - this is under the radar of
                # version_me
                for okey in dup.belongs:
                    owner = dup.belongs[okey]
                    (oid, *att) = okey
                    if len(att) == 2:
                        getattr(owner, att[0]).data[att[1]] = dup
                    else:
                        owner.__dict__[att[0]] = dup
                setattr_func(self, name, value)  #
                # this is on version_me's radar- dups the owning entity if nec
                for okey in self.belongs:
                    owner = self.belongs[okey]
                    (oid, *att) = okey
                    if len(att) == 2:
                        getattr(owner, att[0])[att[1]] = self
                    else:
                        # if att[0]=='category':
                        #   set_trace()
                        setattr(owner, att[0], self)
                    if owner._prev:
                        # dup (old entity) needs to belong to the prev version of owner
                        del dup.belongs[(id(owner), *att)]
                        dup.belongs[(id(owner._prev), *att)] = owner._prev
            else:
                return setattr_func(self, name, value)

        return _version_set_declared_attr

    @version_me
    def _set_declared_attr(self, name, value):
        atts = type(self).attspec[name]
        if atts == "simple":
            pass
        elif atts == "object":
            oldval = self.__dict__.get(name)
            if oldval:
                if oldval == value:
                    # a wash
                    return
                if not self.versioned:
                    del oldval.belongs[(id(self), name)]
                    self.removed_entities.append((name, oldval))
            if isinstance(value, Entity):
                value.belongs[(id(self), name)] = self
        elif atts == "collection":
            if isinstance(value, dict):
                value = CollValue(value, owner=self, owner_key=name)
            if isinstance(value, list):  # convert list of objs to CollValue
                d = {}
                for v in value:
                    d[getattr(v, type(v).mapspec()["key"])] = v
                value = CollValue(d, owner=self, owner_key=name)
        else:
            raise RuntimeError("unknown attspec value '{}'".format(atts))
        self.dirty = 1
        self.__dict__[name] = value

    def __delattr__(self, name):
        del self.__dict__[name]

    def _check_init(self, init):
        for att in type(self).attspec:
            if init[att]:
                self._check_value(att, init[att])

    def _check_value(self, att, value):
        spec = type(self).attspec[att]
        try:
            if spec == "simple":
                if not (
                    isinstance(value, int)
                    or isinstance(value, str)
                    or isinstance(value, float)
                    or isinstance(value, bool)
                    or value is None
                ):
                    raise ArgError(
                        "value for key '{att}' is not a simple scalar".format(att=att)
                    )
            elif spec == "object":
                if not (isinstance(value, Entity) or value is None):
                    raise ArgError(
                        "value for key '{att}' is not an Entity subclass".format(
                            att=att
                        )
                    )
            elif spec == "collection":
                if not (isinstance(value, (dict, list, CollValue))):
                    raise AttributeError(
                        "value for key '{att}' is not a dict,list, or CollValue".format(
                            att=att
                        )
                    )
            else:
                raise ArgError(
                    "unknown attribute type '{type}' for attribute '{att}' in attspec".format(
                        type=spec, att=att
                    )
                )
        except Exception:
            raise

    def dup(self):
        """Duplicate the object, but not too deeply. Mainly for use of the versioning machinery."""
        return type(self)(self)

    def delete(self):
        """Delete self from the database.
        If versioning is active, this will 'deprecate' the entity, but not actually remove it from the db
        """
        if self.versioning_on and self.versioned:
            if type(self).version_count > self._from:
                self._to = type(self).version_count
            else:
                warn(
                    "delete - current version count {vct} is <= entity's _to attribute".format(
                        vct=type(self).version_count
                    )
                )
        else:
            # unlink from other entities
            for okey in self.belongs:
                owner = self.belongs[okey]
                (oid, *att) = okey
                if len(att) == 2:
                    del getattr(owner, att[0])[att[1]]
                else:
                    setattr(owner, att[0], None)

    def dget(self, refresh=False):
        """Update self from the database
        :param boolean refresh: if True, force a retrieval from db;
            if False, retrieve from cache;
            don't disrupt changes already made
        """
        if type(self).object_map:
            return type(self).object_map.get(self, refresh)
        else:
            pass

    def dput(self):
        """Put self to the database.
        This will set the `neoid` property if not yet set.
        """
        if type(self).object_map:
            return type(self).object_map.put(self)
        else:
            pass

    def rm(self, force):
        """Delete self from the database. The object instance survives."""
        if type(self).object_map:
            return type(self).object_map.rm(self, force)
        else:
            pass

    @classmethod
    def attr_doc(cls):
        """Create a docstring for declared attributes on class as configured"""

        def str_for_obj(thing):
            if isinstance(thing, set):
                return "|".join(thing)
            else:
                return thing

        (first, *rest) = cls.__doc__.split("\n")
        if cls.__name__ == "Entity":
            first += " Posesses the following attributes:"
        else:
            first += " Posesses all :class:`Entity` attributes, plus the following:"
        doc = """\
.. py:class:: {cls}

{desc}

""".format(
            desc=first, cls=cls.__name__
        )
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "simple"]:
            doc += """\
  .. py:attribute:: {att}
       :type: simple
""".format(
                att=cls.__name__.lower() + "." + att
            )
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "object"]:
            doc += """\
  .. py:attribute:: {att}
       :type: {obj}
""".format(
                att=cls.__name__.lower() + "." + att,
                obj=str_for_obj(cls.mapspec_["relationship"][att]["end_cls"]),
            )
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "collection"]:
            doc += """\
  .. py:attribute:: {att}
       :type: collection of {obj}
""".format(
                att=cls.__name__.lower() + "." + att,
                obj=str_for_obj(cls.mapspec_["relationship"][att]["end_cls"]),
            )
        doc += "\n"
        return doc


class CollValue(UserDict):
    """A UserDict for housing Entity collection attributes.
    This class contains a hook for recording the Entity that
    owns the value that is being set. The value is marked as belonging
    to the *containing object*, not this collection object.
    It also protects against adding arbitrarily typed elements to the
    collection; it throws unless a value to set is an `Entity`. `__setitem__`
    is instrumented for managing versioning.

    :param owner: `Entity` object of which this collection is an attribute
    :param owner_key: the attribute name of this collection on the owner
    """
    def __init__(self, init=None, *, owner, owner_key):
        self.__dict__["__owner"] = owner
        self.__dict__["__owner_key"] = owner_key
        super().__init__(init)

    @property
    def owner(self):
        """The entity instance of which this collection is an attribute"""
        return self.__dict__["__owner"]

    @property
    def owner_key(self):
        """The attribute name of this collection on the `owner`"""
        return self.__dict__["__owner_key"]

    def version_me(setitem_func):
        def _version_set_collvalue_item(self, name, value):
            if not self.owner.versioning_on:
                return setitem_func(self, name, value)
            if not self.owner.versioned:
                return setitem_func(self, name, value)
            elif (Entity.version_count > self.owner._from) and (self.owner._to is None):
                pass
                # dup becomes the "old" object and self the "new":
                dup = self.owner.dup()
                dup._to = Entity.version_count
                self.owner._from = Entity.version_count
                if self.owner._prev:
                    dup._prev = self.owner._prev
                    self.owner._prev._next = dup
                dup._next = self.owner
                self.owner._prev = dup
                self.owner.neoid = None
                # make the owners own dup, rather than self.owner
                for okey in dup.belongs:
                    owner = dup.belongs[okey]
                    (oid, *att) = okey
                    if len(att) == 2:
                        getattr(owner, att[0]).data[att[1]] = dup
                    else:
                        owner.__dict__[att[0]] = dup
                setitem_func(self, name, value)
                for okey in self.owner.belongs:
                    owner = self.owner.belongs[okey]
                    (oid, *att) = okey
                    if len(att) == 2:
                        getattr(owner, att[0])[
                            att[1]
                        ] = self.owner  # this dups the owning entity if nec
                    else:
                        setattr(owner, att[0], self.owner)
                    if owner._prev:
                        # dup (old entity) needs to belong to the prev version of owner
                        del dup.belongs[(id(owner), *att)]
                        dup.belongs[(id(owner._prev), *att)] = owner._prev
            else:
                return setitem_func(self, name, value)

        return _version_set_collvalue_item

    @version_me
    def __setitem__(self, name, value):
        if not isinstance(value, Entity):
            raise ArgError(
                "a collection-valued attribute can only accept Entity members, not '{tipe}'s".format(
                    tipe=type(value)
                )
            )
        if name in self:
            oldval = self.data.get(name)
            if oldval:
                if oldval == value:
                    # a wash
                    return
                if not self.owner.versioned:
                    del oldval.belongs[(id(self.owner), self.owner_key, name)]
                    self.owner.removed_entities.append((self.owner_key, oldval))
        value.belongs[(id(self.owner), self.owner_key, name)] = self.owner
        # smudge the owner
        self.owner.dirty = 1
        self.data[name] = value
        return

    def __getitem__(self, name):
        if name not in self.data:
            return
        if self.data[name].dirty < 0:
            self.data[name].dget()
        return self.data[name]

    def __delitem__(self, name):
        self[name] == None  # trigger __setitem__
        super().__delitem__(name)
        return



"""
bento_meta.util.cypher.functions

Representations of Cypher functions
"""
from string import Template

# cypher functions


class Func(object):
    template = Template("func(${slot1})")
    joiner = ','
    As = ""

    @staticmethod
    def context(arg):
        return _return(arg)

    def __init__(self, arg):
        self.arg = arg

    def __str__(self):
        slot = ""
        if type(self.arg) == list:
            slot = self.joiner.join([self.context(a) for a in self.arg])
        else:
            slot = self.context(self.arg)
        if self.As:
            return self.template.substitute(slot1=slot)+" as "+self.As
        else:
            return self.template.substitute(slot1=slot)


class count(Func):
    template = Template("count($slot1)")


class exists(Func):
    template = Template("exists($slot1)")

class labels(Func):
    template = Template("labels($slot1)")


class Not(Func):
    template = Template("NOT $slot1")

    @staticmethod
    def context(arg):
        return _condition(arg)


class And(Func):
    template = Template("$slot1")
    joiner = " AND "

    @staticmethod
    def context(arg):
        return _condition(arg)

    def __init__(self, *args):
        self.arg = list(args)
        super().__init__(self.arg)


class Or(Func):
    template = Template("$slot1")
    joiner = " OR "

    @staticmethod
    def context(arg):
        return _condition(arg)

    def __init__(self, *args):
        self.arg = list(args)
        super().__init__(self.arg)


class group(Func):
    template = Template("($slot1)")
    joiner = " "


class is_null(Func):
    template = Template("$slot1 IS NULL")


class is_not_null(Func):
    template = Template("$slot1 IS NOT NULL")

# rendering contexts


def _pattern(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.pattern()


def _condition(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.condition()


def _return(ent):
    if isinstance(ent, (str, Func)):
        return str(ent)
    return ent.Return()



"""
ToolsMDB: subclass of 'WriteableMDB' to support interactions with the MDB.
"""

from pathlib import Path
import csv
from typing import List
import logging
from logging.config import fileConfig

import spacy
from nanoid import generate
from bento_meta.entity import Entity
from bento_meta.mdb import read_txn, read_txn_data, read_txn_value
from bento_meta.mdb.writeable import WriteableMDB, write_txn
from bento_meta.objects import Concept, Predicate, Property, Term
from bento_meta.util.cypher.clauses import Match, Return, Statement
from bento_meta.util.cypher.entities import N, VarLenR, NoDirT

# logging stuff
log_ini_path = Path(__file__).parents[2].joinpath("logs/log.ini")
log_file_path = Path(__file__).parents[2].joinpath(f"logs/{__name__}.log")
fileConfig(log_ini_path, defaults={'logfilename': log_file_path.as_posix()})
logger = logging.getLogger(__name__)


class EntNotFoundError(Exception):
    """Raised when an entity is not found and shouldn't be created"""


def get_entity_type(entity: Entity):
    """returns type of entity"""
    if entity.__class__.__name__.lower() == "edge":
        return "relationship"
    return entity.__class__.__name__.lower()


class ToolsMDB(WriteableMDB):
    """Adds mdb-tools to WriteableMDB"""

    def __init__(self, uri, user, password):
        WriteableMDB.__init__(self, uri, user, password)

    # FIXME: this should interpolate cypher $parameters, not the actual
    # values
    # (use utils.cypher...)
    def get_entity_attrs(self, entity: Entity, output_str: bool = True):
        """
        Returns attributes as str or dict for given entity.

        Param output_as_str parm defaults to True and returns a string
        that is usable in Cypher query as node properties.

        If output_as_str = False, return dict of attributes instead,
        which is used as params of function with write_txn decorator.
        """
        attr_dict = {}
        for key, val in vars(entity).items():
            if val and val is not None and key != "pvt":
                attr_dict[key] = val

        attr_str = "{"
        for key, val in attr_dict.items():
            attr_str += f"{key}: '{val}', "
        attr_str = attr_str[:-2]
        attr_str += "}"

        if not output_str:
            return attr_dict
        return attr_str

    @write_txn
    def detach_delete_entity(self, entity: Entity) -> tuple:
        """
        Remove given Entity node from the database.

        Accepts the following bento-meta Entities:
            Concept, Node, Predicate, Property, Edge, Term
        """
        entity_type = get_entity_type(entity)
        entity_attr_str = self.get_entity_attrs(entity)
        entity_attr_dict = self.get_entity_attrs(entity, output_str=False)

        qry = f"MATCH (e:{entity_type} " f"{entity_attr_str}) DETACH DELETE e"

        logging.info(f"Removing {entity_type} node with with properties: {entity_attr_str}")
        return (qry, entity_attr_dict)

    @read_txn_value
    def get_entity_count(self, entity: Entity) -> tuple:
        """
        Returns count of given entity (w/ its properties) found in MDB.

        If count = 0, entity with given properties not found in MDB.
        If count = 1, entity with given properties is unique in MDB
        If count > 1, more properties needed to uniquely id entity in MDB.
        """
        # if not (term.origin_name and term.value):
        #     raise RuntimeError("arg 'term' must have both origin_name and value")
        entity_type = get_entity_type(entity)
        if entity.nanoid:
            entity_attr_str = "{nanoid:$nanoid}"
            entity_attr_dict = {"nanoid": entity.nanoid}
        else:
            entity_attr_str = self.get_entity_attrs(entity)
            entity_attr_dict = self.get_entity_attrs(entity, output_str=False)

        qry = (
            f"MATCH (e:{entity_type} {entity_attr_str}) "
            "RETURN COUNT(e) as entity_count"
        )

        return (qry, entity_attr_dict, "entity_count")

    @read_txn_value
    def get_triple_count(
        self,
        src_entity: Entity,
        dst_entity: Entity,
        relationship: str,
    ) -> tuple:
        """
        Returns count of given triple, (src)-[rel]->(dst), found in MDB.

        If count = 0, triple with given properties not found in MDB.
        If count = 1, triple with given properties is unique in MDB
        If count > 1, more properties needed to uniquely id triple in MDB.
        """
        src_entity_type = get_entity_type(src_entity)
        dst_entity_type = get_entity_type(dst_entity)

        qry = (
            "MATCH (s:{src_type} {{nanoid:$src_nano}}) "
            "-[:{relationship}]->"
            "(d:{dst_type} {{nanoid: $dst_nano}}) "
            "RETURN COUNT(*) as triple_count".format(
                src_type=src_entity_type,
                dst_type=dst_entity_type,
                relationship=relationship,
            )
        )

        parms = {
            "src_nano": src_entity.nanoid,
            "dst_nano": dst_entity.nanoid,
        }

        return (qry, parms, "triple_count")

    @write_txn
    def create_entity(
        self,
        entity: Entity,
        _commit=None,
    ) -> tuple:
        """Adds given Entity node to database"""
        if not entity.nanoid:
            logging.error("entity has no nanoid")
            raise RuntimeError(
                "Entity needs a nanoid before creating - please set valid entity nanoid. "
                "entity_name.nanoid = mdb_name.get_or_make_nano(entity_name) AFTER "
                "declaring some other entity properties is a good way to do this."
            )
        if _commit:
            entity._commit = _commit
        entity_type = get_entity_type(entity)
        entity_attr_str = self.get_entity_attrs(entity)
        entity_attr_dict = self.get_entity_attrs(entity, output_str=False)

        qry = f"MERGE (e:{entity_type} {entity_attr_str})"

        logging.info(f"Creating new {entity_type} node with properties: {entity_attr_str}")
        return (qry, entity_attr_dict)

    @read_txn_value
    def get_concepts(self, entity: Entity):
        """Returns list of concepts represented by given entity"""
        # if not (term.origin_name and term.value):
        #     raise RuntimeError("arg 'term' must have both origin_name and value")
        entity_type = get_entity_type(entity)

        if entity.nanoid:
            entity_attr_str = "{nanoid:$nanoid}"
            entity_attr_dict = {"nanoid": entity.nanoid}
        else:
            entity_attr_str = self.get_entity_attrs(entity)
            entity_attr_dict = self.get_entity_attrs(entity, output_str=False)

        if entity_type == "term":
            qry = (
                f"MATCH (e:{entity_type} {entity_attr_str}) "
                "-[:represents]->(c:concept) RETURN c.nanoid AS concept"
            )
        else:
            qry = (
                f"MATCH (e:{entity_type} {entity_attr_str}) "
                "-[:has_concept]->(c:concept) RETURN c.nanoid AS concept"
            )
        return (qry, entity_attr_dict, "concept")

    @write_txn
    def create_relationship(
        self,
        src_entity: Entity,
        dst_entity: Entity,
        relationship: str,
        _commit=None,
    ):
        """Adds relationship between given entities in MDB"""
        # gets number of entities in MDB matching given entity (w/ given properties)
        ent_1_count = self.get_entity_count(src_entity)[0]
        ent_2_count = self.get_entity_count(dst_entity)[0]
        # at least one of the given entities doesn't uniquely id a node in the MDB.
        if ent_1_count > 1 or ent_2_count > 1:
            logging.error("Given attributes don't uniquely identify MDB entities.")
            raise RuntimeError(
                "Given entities must uniquely identify nodes in the MDB. Please add "
                "necessary properties to the entity so that it can be uniquely identified."
            )
        # at least one of given entities not found and shouldn't be added
        if ent_1_count < 1 or ent_2_count < 1:
            logging.error("One or more of the given entities aren't in the MDB.")
            raise RuntimeError("One or more of the given entities aren't in the MDB.")
        src_entity_type = get_entity_type(src_entity)
        dst_entity_type = get_entity_type(dst_entity)

        src_attr_str = self.get_entity_attrs(src_entity)
        dst_attr_str = self.get_entity_attrs(dst_entity)
        _commit_str = ""
        if _commit:
            _commit_str = f"{{_commit:'{_commit}'}}"

        # check if triple (ent1)-[relationship]-(ent2) already exists
        trip_count = self.get_triple_count(
            src_entity=src_entity,
            dst_entity=dst_entity,
            relationship=relationship,
        )[0]

        if trip_count > 1:
            logging.error("Relationship exists more than once")
            raise RuntimeError(
                "This relationship exists more than once, check the MDB "
                "instance as this shouldn't be the case."
            )
        elif trip_count == 1:
            logging.warning(
                f"{relationship} relationship already exists between src {src_entity_type} with "
                f"properties: {src_attr_str} to dst {dst_entity_type} with properties: {dst_attr_str}"
            )
            # way to exit out of decorator? maybe can fix w/ upcoming cypher.utils integration
            # for now removing commit str before merging (which won't create an extra relationship)
            qry = (
                "MATCH (s:{src_type} {{nanoid:$src_nano}}), "
                "(d:{dst_type} {{nanoid: $dst_nano}}) "
                "MERGE (s)-[:{relationship}]->(d)".format(
                    src_type=src_entity_type,
                    dst_type=dst_entity_type,
                    relationship=relationship,
                )
            )
            parms = {
                "src_nano": src_entity.nanoid,
                "dst_nano": dst_entity.nanoid,
            }
            return (qry, parms)

        qry = (
            "MATCH (s:{src_type} {{nanoid:$src_nano}}), "
            "(d:{dst_type} {{nanoid: $dst_nano}}) "
            "MERGE (s)-[:{relationship} {commit}]->(d)".format(
                src_type=src_entity_type,
                dst_type=dst_entity_type,
                relationship=relationship,
                commit=_commit_str,
            )
        )

        parms = {
            "src_nano": src_entity.nanoid,
            "dst_nano": dst_entity.nanoid,
        }
        logging.info(
            f"Adding {relationship} relationship between src {src_entity_type} with "
            f"properties: {src_attr_str} to dst {dst_entity_type} with properties: {dst_attr_str}"
        )
        return (qry, parms)

    def link_synonyms(
        self,
        entity_1: Entity,
        entity_2: Entity,
        add_missing_ent_1: bool = False,
        add_missing_ent_2: bool = False,
        _commit=None,
    ):
        """
        Link two synonymous entities in the MDB via a Concept node.

        This function takes two synonymous Entities (as determined by user/SME) as
        bento-meta objects connects them to a Concept node via a 'represents' relationship.

        If one or both doesn't exist in the MDB and add_missing_ent is True they will be added.

        If one or both doesn't uniquely identify a node in the MDB will give error.

        If _commit is set (to a string), the _commit property of any node created is set to
        this value.
        """
        # gets number of entities in MDB matching given entity (w/ given properties)
        ent_1_count = self.get_entity_count(entity_1)[0]
        ent_2_count = self.get_entity_count(entity_2)[0]
        # at least one of the given entities doesn't uniquely id a node in the MDB.
        if ent_1_count > 1 or ent_2_count > 1:
            logging.error("Given entities don't uniquely identify MDB entities")
            raise RuntimeError(
                "Given entities must uniquely identify nodes in the MDB. Please add "
                "necessary properties to the entity so that it can be uniquely identified."
            )
        # entity 1 not found and shouldn't be added
        if ent_1_count < 1 and not add_missing_ent_1:
            logging.error("Entity 1 isn't in the MDB and add_missing_ent_1 is False.")
            raise EntNotFoundError(
                "Entity 1 isn't in the MDB and add_missing_ent_1 is False."
                "Please add the missing entity to the MDB or set add_missing_ent_1 to True."
            )
        # entity 2 not found and shouldn't be added
        if ent_2_count < 1 and not add_missing_ent_2:
            logging.error("Entity 2 isn't in the MDB and add_missing_ent_2 is False.")
            raise EntNotFoundError(
                "Entity 2 isn't in the MDB and add_missing_ent_2 is False."
                "Please add the missing entity to the MDB or set add_missing_ent_2 to True."
            )
        # entity 1 not found and should be added
        if not ent_1_count and add_missing_ent_1:
            self.create_entity(entity_1, _commit=_commit)
        # entity 2 not found and should be added
        if not ent_2_count and add_missing_ent_2:
            self.create_entity(entity_2, _commit=_commit)
        # get any existing concepts and create new concept if none found
        ent_1_concepts = self.get_concepts(entity_1)
        ent_2_concepts = self.get_concepts(entity_2)
        if ent_1_concepts:
            concept = Concept({"nanoid": ent_1_concepts[0]})
        elif ent_2_concepts:
            concept = Concept({"nanoid": ent_2_concepts[0]})
        else:
            concept = Concept({"nanoid": self.make_nano()})
            self.create_entity(concept, _commit=_commit)
        # entities are already connected by a concept
        if not set(ent_1_concepts).isdisjoint(set(ent_2_concepts)):
            concept = set(ent_1_concepts).intersection(set(ent_2_concepts))
            logging.warning(f"Both entities are already connected via Concept {list(concept)[0]}")
            return
        # create specified relationship between each entity and a concept
        if get_entity_type(entity_1) == "term":
            self.create_relationship(entity_1, concept, "represents", _commit=_commit)
            self.create_relationship(entity_2, concept, "represents", _commit=_commit)
        else:
            self.create_relationship(entity_1, concept, "has_concept", _commit=_commit)
            self.create_relationship(entity_2, concept, "has_concept", _commit=_commit)

    def make_nano(self):
        """Generates valid nanoid"""
        return generate(
            size=6,
            alphabet="abcdefghijkmnopqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ0123456789",
        )

    def _get_prop_nano(self, prop: Entity, node_handle: str):
        """
        Takes Property and Node entities and returns Property nanoid.

        This is used to help uniquely identify a property node in the MDB,
        which requires a node connected via the has_property relationship.
        """
        qry = (
            "MATCH (p:property {handle:$prop_handle, model: $prop_model})"
            "<-[:has_property]-(n:node {handle: $node_handle}) "
            "RETURN p.nanoid as prop_nano"
        )
        parms = {
            "prop_handle": prop.handle,
            "prop_model": prop.model,
            "node_handle": node_handle,
        }
        return (qry, parms, "prop_nano")

    def _get_edge_nano(self, edge: Entity, src_handle: str, dst_handle: str):
        """
        Takes Edge, src node handle, and dst Node handle and returns Edge nanoid.

        This is used to help uniquely identify an edge node in the MDB,
        which requires nodes connected via the has_src and has_dst relationships.
        """
        qry = (
            "MATCH (s:node {handle: $src_handle})<-[:has_src]-"
            "(r:relationship {handle: $edge_handle, model: $edge_model})-"
            "[:has_dst]->(d:node {handle: $dst_handle}) "
            "RETURN r.nanoid as edge_nano"
        )
        parms = {
            "edge_handle": edge.handle,
            "edge_model": edge.model,
            "src_handle": src_handle,
            "dst_handle": dst_handle,
        }
        return (qry, parms, "edge_nano")

    @read_txn_value
    def get_entity_nano(
        self, entity: Entity, extra_handle_1: str = "", extra_handle_2: str = ""
    ) -> tuple:
        """
        Takes an entity and returns its nanoid. If entity requires handles of connected nodes
        for unique identification (Property or Edge), extra_handle_1 and _2 hold these as str.

        Note: If an entity exists in the MDB with the given properties, but doesn't have an
        assigned nanoid for some reason, returns [None] instead of [].
        """
        ent_type = get_entity_type(entity)
        if ent_type == "property":
            if not extra_handle_1:
                logging.error("Property entity doesn't have connected node handle")
                raise RuntimeError(
                    "Property entities require the handle of a node connected via 'has_property' "
                    "for unique identification. Set 'extra_handle_1' to that node handle str."
                )
            return self._get_prop_nano(prop=entity, node_handle=extra_handle_1)
        if ent_type == "relationship":
            if not extra_handle_1 or not extra_handle_2:
                logging.error("Edge entity doesn't have src or dst node handle")
                raise RuntimeError(
                    "Edge entities require the handles of a node connected via "
                    "'has_src' relationship and of a node connected via 'has_dst' "
                    "relationship for unique identification. Set 'extra_handle_1' to "
                    "the src handle and 'extra_handle_2' to the dst handle"
                )
            return self._get_edge_nano(
                edge=entity, src_handle=extra_handle_1, dst_handle=extra_handle_2
            )

        ent_count = self.get_entity_count(entity)[0]
        if ent_count > 1:
            logging.error("Given entities don't uniquely identify MDB entities")
            raise RuntimeError(
                "Given entities must uniquely identify nodes in the MDB. Please add "
                "necesary properties to the entity so that it can be uniquely identified."
            )
        # if ent_count < 1:
        #     raise RuntimeError("Given entity wasn't found in the MDB")

        ent_attr_str = self.get_entity_attrs(entity)
        ent_attr_dict = self.get_entity_attrs(entity, output_str=False)

        qry = f"MATCH (e:{ent_type} {ent_attr_str}) RETURN e.nanoid as ent_nano"

        return (qry, ent_attr_dict, "ent_nano")

    def get_or_make_nano(
        self, entity: Entity, extra_handle_1: str = "", extra_handle_2: str = ""
    ) -> str:
        """Obtains existing entity's nanoid or creates one for new entity."""
        nano_list = self.get_entity_nano(entity, extra_handle_1, extra_handle_2)
        if len(nano_list) > 1:
            logging.error("Given entity doesn't uniquely identify MDB entities")
            raise RuntimeError(
                "More than one entity exists with these properties. Please "
                "add more properties of the desired entity to uniquely ID."
            )
        elif nano_list and not nano_list[0]:
            logging.error("Entity exists but doesn't have nanoid")
            raise RuntimeError(
                "An entity with these properties exists in the MDB but doesn't "
                "have an assigned nanoid for some reason."
            )
        elif nano_list:
            nano = nano_list[0]
            return nano
        nano = self.make_nano()
        return nano

    @read_txn_value
    def get_term_nanos(self, concept: Concept):
        """Returns list of term nanoids representing given concept"""
        if not concept.nanoid:
            logging.error("Concept doesn't have a nanoid")
            raise RuntimeError("arg 'concept' must have nanoid")
        entity_attr_str = self.get_entity_attrs(concept)
        entity_attr_dict = self.get_entity_attrs(concept, output_str=False)
        qry = (
            f"MATCH (t:term)-[:represents]->(c:concept {entity_attr_str}) "
            "RETURN t.nanoid AS term_nano"
        )
        return (qry, entity_attr_dict, "term_nano")

    @read_txn_value
    def get_predicate_nanos(self, concept: Concept):
        """Returns list of predicate nanoids with relationship to given concept"""
        if not concept.nanoid:
            logging.error("Concept doesn't have a nanoid")
            raise RuntimeError("arg 'concept' must have nanoid")
        entity_attr_str = self.get_entity_attrs(concept)
        entity_attr_dict = self.get_entity_attrs(concept, output_str=False)
        qry = (
            f"MATCH (p:predicate)-[r]->(c:concept {entity_attr_str}) "
            "RETURN p.nanoid AS pred_nano"
        )
        return (qry, entity_attr_dict, "pred_nano")

    @read_txn_value
    def get_predicate_relationship(self, concept: Concept, predicate: Predicate):
        """Returns relationship type between given concept and predicate"""
        if not concept.nanoid or not predicate.nanoid:
            logging.error("Concept or Predicate don't have a nanoid")
            raise RuntimeError("args 'concept' and 'predicate' must have nanoid")
        qry = (
            "MATCH (p:predicate {nanoid: $pred_nano})-[r]->(c:concept {nanoid: $con_nano}) "
            "RETURN TYPE(r) as rel_type"
        )
        parms = {"pred_nano": predicate.nanoid, "con_nano": concept.nanoid}
        return (qry, parms, "rel_type")

    def link_concepts_to_predicate(
        self,
        concept_1: Concept,
        concept_2: Concept,
        predicate_handle: str = "exactMatch",
        _commit=None,
    ) -> None:
        """
        Links two synonymous Concepts via a Predicate

        This function takes two synonymous Concepts as objects and links
        them via a Predicate node and has_subject and has_object relationships.
        """
        if not (concept_1.nanoid and concept_2.nanoid):
            logging.error("Concepts don't have a nanoid")
            raise RuntimeError("args 'concept_1' and 'concept_2' must have nanoid")
        valid_predicate_handles = [
            "exactMatch",
            "closeMatch",
            "broader",
            "narrower",
            "related",
        ]
        if predicate_handle not in valid_predicate_handles:
            logging.error(f"Predicate handle: {predicate_handle} not a valid predicate handle")
            raise RuntimeError(
                f"'handle' key must be one the following: {valid_predicate_handles}"
            )
        # create predicate
        new_predicate = Predicate(
            {"handle": predicate_handle, "nanoid": self.make_nano()}
        )
        self.create_entity(new_predicate, _commit=_commit)
        # link concepts to predicate via subject & object relationships
        self.create_relationship(
            new_predicate, concept_1, "has_subject", _commit=_commit
        )
        self.create_relationship(
            new_predicate, concept_2, "has_object", _commit=_commit
        )

    def merge_two_concepts(
        self, concept_1: Concept, concept_2: Concept, _commit=None
    ) -> None:
        """
        Combine two synonymous Concepts into a single Concept.

        This function takes two synonymous Concept as bento-meta objects and
        merges them into a single Concept along with any connected Terms and Predicates.
        """
        if not (concept_1.nanoid and concept_2.nanoid):
            logging.error("Concept doesn't have a nanoid")
            raise RuntimeError("args 'concept_1' and 'concept_2' must have nanoid")
        # get list of terms connected to concept 2
        c2_term_nanos = self.get_term_nanos(concept_2)
        c2_terms = []
        for nano in c2_term_nanos:
            c2_terms.append(Term({"nanoid": nano}))
        # get list of predicates w/ relationship type connected to concept 2
        c2_pred_nanos = self.get_predicate_nanos(concept_2)
        c2_pred_rels = []
        for nano in c2_pred_nanos:
            pred = Predicate({"nanoid": nano})
            rel = self.get_predicate_relationship(concept_2, pred)[0]
            c2_pred_rels.append([pred, rel])
        # delete concept 2
        self.detach_delete_entity(concept_2)
        # connect terms from deleted (c2) to remaining concept (c1)
        for term in c2_terms:
            self.create_relationship(term, concept_1, "represents", _commit=_commit)
        # connect predicates from deleted (c2) to remaining concept (c1)
        for pred_rel in c2_pred_rels:
            c2_pred = pred_rel[0]
            c2_rel = pred_rel[1]
            self.create_relationship(c2_pred, concept_1, c2_rel, _commit=_commit)

    @read_txn
    def _get_all_terms(self):
        """Gets value, origin_name, and nanoid for all terms in database."""
        qry = (
            "MATCH (t:term) "
            "RETURN t.value AS term_val, t.origin_name AS term_origin, t.nanoid as term_nano"
        )
        return (qry, {})

    def get_term_synonyms(self, term: Term, threshhold: float = 0.8) -> List[dict]:
        """Returns list of dicts representing Term nodes synonymous to given Term"""
        if not (term.origin_name and term.value):
            logging.error("Term doesn't have origin_name and value")
            raise RuntimeError("arg 'term' must have both origin_name and value")
        # load spaCy NER model
        nlp = spacy.load("en_ner_bionlp13cg_md")
        # get result with all term values, origin_names, and nanoids in the database
        all_terms_result = self._get_all_terms()
        # get likely synonyms
        synonyms = []
        for record in all_terms_result:
            # calculate similarity between each Term and input Term
            term_1 = nlp(term.value)
            term_2 = nlp(str(record["term_val"]))  # type: ignore
            similarity_score = term_1.similarity(term_2)
            # if similarity threshold met, add to list of potential synonyms
            if similarity_score >= threshhold:
                synonym = {
                    "value": record["term_val"],  # type: ignore
                    "origin_name": record["term_origin"],  # type: ignore
                    "nanoid": record["term_nano"],  # type: ignore
                    "similarity": similarity_score,
                    "valid_synonym": 0,  # mark 1 if synonym when uploading later
                }
                synonyms.append(synonym)
        synonyms_sorted = sorted(synonyms, key=lambda d: d["similarity"], reverse=True)
        return synonyms_sorted

    def potential_synonyms_to_csv(
        self, input_data: List[dict], output_path: str
    ) -> None:
        """Given a list of synonymous Terms as dicts, outputs to CSV file at given output path"""
        with open(output_path, "w", encoding="utf8", newline="") as output_file:
            dict_writer = csv.DictWriter(output_file, fieldnames=input_data[0].keys())
            dict_writer.writeheader()
            dict_writer.writerows(input_data)

    def link_term_synonyms_csv(self, term: Term, csv_path: str, _commit=None) -> None:
        """Given a CSV of syonymous Terms, links each via a Concept node to given Term"""
        with open(csv_path, encoding="UTF-8") as csvfile:
            synonym_reader = csv.reader(csvfile)
            for line in synonym_reader:
                if line[3] == "1":
                    synonym = Term()
                    synonym.value = line[0]
                    synonym.origin_name = line[1]
                    self.link_synonyms(term, synonym, _commit=_commit)

    @read_txn_data
    def get_property_synonyms(self, prop: Property):
        """
        Returns list of properties linked by concept to given property
        or to synonym of given property
        """
        if not prop.nanoid:
            logging.error("Property doesn't have a nanoid")
            raise RuntimeError("property entity needs a nanoid")
        p_attrs = self.get_entity_attrs(prop, output_str=False)
        prop_1 = N(label="property", props=p_attrs)
        prop_2 = N(label="property")
        vl_rel = VarLenR(Type="has_concept")
        trip = NoDirT(prop_1, vl_rel, prop_2)

        stmt = Statement(
            Match(trip),
            Return(f"distinct({prop_2.Return()})"),
            use_params=True
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms)



"""
bento_meta.object_map
=====================

This module contains :class:`ObjectMap`, a class which provides the 
machinery for mapping bento_meta objects to a Bento Metamodel Database
in Neo4j. Mostly not for human consumption. The ObjectMap:

* interprets the attribute specification (attspec) and map
  specification (mapspec) associated with :class:`Entity` subclasses
* provides the :meth:`get` and :meth:`put` methods to subclasses, that
  enable them to get and put themselves to the database
* generates appropriate `Cypher <https://neo4j.com/docs/cypher-manual/current/>`_ queries to do gets and puts

One ObjectMap instance should be generated for each Entity subclass (see, e.g., 
:class:`bento_meta.model.Model`)

"""
import re
import sys
sys.path.append("..")
from neo4j import BoltDriver, Neo4jDriver
from bento_meta.entity import *
from bento_meta.objects import *
# from pdb import set_trace


class ObjectMap(object):
    cache = {}

    def __init__(self, *, cls=None, drv=None):
        if not cls:
            raise ArgError("arg cls= is required")
        self.cls = cls
        if drv:
            if isinstance(drv, (Neo4jDriver, BoltDriver)):
                self.drv = drv
            else:
                raise ArgError(
                    "drv= arg must be Neo4jDriver or BoltDriver (returned from GraphDatabase.driver())"
                )
        self.maps = {}

    @classmethod
    def clear_cache(cls):
        cls.cache = {}

    @classmethod
    def cls_by_label(cls, lbl):
        if not hasattr(cls, "_clsxlbl"):
            cls._clsxlbl = {}
            for o in (Node, Edge, Property, ValueSet, Term, Concept, Origin, Tag):
                cls._clsxlbl[o.mapspec()["label"]] = o
        return cls._clsxlbl.get(lbl)

    @classmethod
    def keys_by_cls_and_reln(cls, qcls, reln):
        if not hasattr(cls, "_keysxcls"):
            cls._keysxcls = {}
            for o in (Node, Edge, Property, ValueSet, Term, Concept, Predicate, Origin, Tag):
                for oatt in [x for x in o.attspec if o.attspec[x] == "object"]:
                    r = o.mapspec()["relationship"][oatt]["rel"]
                    r = re.match("[:<>]*([a-zA-Z_]+)[:<>]*", r).group(1)
                    cls._keysxcls[(o.__name__, r)] = (oatt, None)
                for catt in [x for x in o.attspec if o.attspec[x] == "collection"]:
                    r = o.mapspec()["relationship"][catt]["rel"]
                    r = re.match("[:<>]*([a-zA-Z_]+)[:<>]*", r).group(1)
                    cls._keysxcls[(o.__name__, r)] = (catt, o.mapspec()["key"])
        return cls._keysxcls.get((qcls.__name__, reln))

    @classmethod
    def _quote_val(cls, value, single=None):  # double quote unless single is set
        if value is None:
            return
        if isinstance(value, (int, float)):
            return value  # no quote
        else:
            if single:
                return "'{val}'".format(val=value)  # quote
            else:
                return '"{val}"'.format(val=value)  # quote

    def get_by_id(self, obj, id, refresh=False):
        """Get an entity given an id attribute value (not the Neo4j id)"""
        neoid = None

        if self.drv is None:
            raise ArgError("get_by_id() requires Neo4j driver instance")

        with self.drv.session() as session:
            result = session.run(self.get_by_id_q(), {"id": id})
            rec = (
                result.single()
            )  # should be unique - this call will warn if there are more than one

            if rec is not None:
                neoid = rec["id(n)"]

        if neoid is not None:
            obj.neoid = neoid
            return self.get(obj, refresh=True)
        else:
            return

    def get_by_node_nanoid(self, obj, nanoid, refresh=False):
        """PROTOTYPE
        Get an entity given an id attribute value (not the Neo4j id)
        """

        neo4jid = None

        if not self.drv:
            raise ArgError("get_by_id() requires Neo4j driver instance")

        with self.drv.session() as session:
            result = session.run(self.get_by_node_nanoid_q(), {"nanoid": nanoid})
            rec = (
                result.single()
            )  # should be unique - this call will warn if there are more than one
            if rec is not None:
                neo4jid = rec["id(n)"]
           
        if neo4jid is None:
            obj.neoid = neo4jid
            return self.get(obj, refresh=True)
        else:
            return

    def get(self, obj, refresh=False):
        """Get the data for an object instance from the db and load the instance with it"""
        if not self.drv:
            raise ArgError("get() requires Neo4j driver instance")

        if refresh:
            pass
        elif (obj.neoid in ObjectMap.cache) and (ObjectMap.cache[obj.neoid].dirty >= 0):
            return obj

        with self.drv.session() as session:
            result = session.run(self.get_q(obj))
            rec = result.single()
            if not rec:
                raise RuntimeError(
                    "object with id {neoid} not found in db".format(neoid=obj.neoid)
                )

        if obj.neoid not in ObjectMap.cache:
            ObjectMap.cache[obj.neoid] = obj

        with self.drv.session() as session:
            for att in self.cls.mapspec()["relationship"]:
                result = session.run(self.get_attr_q(obj, att))
                values = {}
                first_val = None
                for rec in result:
                    o = ObjectMap.cache.get(rec["a"].id)
                    if o:
                        if not first_val:
                            first_val = o
                        values[getattr(o, type(o).mapspec()["key"])] = o
                    else:
                        c = None
                        for l in rec["a"].labels:
                            c = ObjectMap.cls_by_label(l)
                            if c:
                                break
                        if not c:
                            raise RuntimeError(
                                "node labels {lbls} have no associated class in the object model".format(
                                    lbls=rec["a"].labels
                                )
                            )
                        o = c(rec["a"])
                        o.dirty = -1
                        ObjectMap.cache[o.neoid] = o
                        if not first_val:
                            first_val = o
                        values[getattr(o, type(o).mapspec()["key"])] = o
                if self.cls.attspec[att] == "object" and len(values) > 1:
                    warn(
                        "expected one node for attribute {att} on class {cls}, but got {n}; using first one".format(
                            att=att, cls=self.cls.__name__, n=len(values)
                        )
                    )
                if self.cls.attspec[att] == "object":
                    setattr(obj, att, first_val)
                elif self.cls.attspec[att] == "collection":
                    setattr(obj, att, values)
                else:
                    raise RuntimeError(
                        "attribute '{att}' has unknown attribute type '{atype}'".format(
                            att=att, atype=self.cls.attspec[att]
                        )
                    )

        obj.clear_removed_entities()
        obj.dirty = 0
        return obj

    def put(self, obj):
        """Put the object instance's attributes to the mapped data node in the database"""
        if not self.drv:
            raise ArgError("put() requires Neo4j driver instance")
            pass
        with self.drv.session() as session:
            result = None
            with session.begin_transaction() as tx:
                for qry in self.put_q(obj):
                    result = tx.run(qry)
                obj.neoid = result.single().value("id(n)")
                if obj.neoid is None:
                    raise RuntimeError(
                        "no neo4j id retrived on put for obj '{name}'".format(
                            name=getattr(obj, self.cls.mapspec()["key"])
                        )
                    )
                for att in self.cls.mapspec()["relationship"]:
                    values = getattr(obj, att)
                    if not values:
                        continue
                    if isinstance(values, CollValue):
                        items = values.values()
                    else:
                        items = [values]
                    for val in items:
                        if val.neoid is not None:
                            continue
                        # put val as a node
                        for qry in ObjectMap(cls=type(val), drv=self.drv).put_q(val):
                            result = tx.run(qry)
                        val.neoid = result.single().value("id(n)")
                        if val.neoid is None:
                            raise RuntimeError(
                                "no neo4j id retrived on put for obj '{name}'".format(
                                    name=val[type(val).mapspec()["key"]]
                                )
                            )
                        val.dirty = 1
                        ObjectMap.cache[val.neoid] = val
                    for qry in self.put_attr_q(obj, att, values):
                        tx.run(qry)
                    # drop removed entities here
                    while obj.removed_entities:
                        ent = obj.removed_entities.pop()
                        self.drop(obj, *ent, tx)
        ObjectMap.cache[obj.neoid] = obj
        obj.dirty = 0
        return obj

    def rm(self, obj, force=False):
        """'Delete' the object's mapped node from the database"""
        if not self.drv:
            raise ArgError("rm() requires Neo4j driver instance")
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        with self.drv.session() as session:
            result = session.run(self.rm_q(obj, force))
            s = result.single()
            if s is None:
                warn("rm() - corresponding db node not found")
            else:
                return s.value()

    def add(self, obj, att, tgt):
        """Create a link between an object instance and a target object in the database.
        This represents adding an object-valued attribute to the object.
        """
        if not self.drv:
            raise ArgError("add() requires Neo4j driver instance")
        with self.drv.session() as session:
            for qry in self.put_attr_q(obj, att, tgt):
                result = session.run(qry)
            tgt_id = result.single().value()
            if tgt_id is None:
                warn("add() - corresponding db node not found")
            return tgt_id

    def drop(self, obj, att, tgt, tx=None):
        """Remove an existing link between an object instance and a target object in the database.
        This represents dropping an object-valued attribute from the object.
        """
        if not self.drv:
            raise ArgError("rm() requires Neo4j driver instance")
        # if the tgt is not in the database, then dropping it is a no-op:
        if not tgt.neoid:
            return

        if tx:
            result = None
            for qry in self.rm_attr_q(obj, att, tgt):
                result = tx.run(qry)
            s = result.single()
            if s is None:
                warn("drop() - corresponding target db node not found")
            else:
                return s.value()
        else:
            with self.drv.session() as session:
                result = None
                for qry in self.rm_attr_q(obj, att, tgt):
                    result = session.run(qry)
                s = result.single()
                if s is None:
                    warn("drop() - corresponding target db node not found")
                else:
                    return s.value()

    def get_owners(self, obj):
        """Get the nodes which are linked to the object instance (the owners of the object)"""
        if not self.drv:
            raise ArgError("get_owners() requires Neo4j driver instance")
        ret = []
        with self.drv.session() as session:
            result = session.run(self.get_owners_q(obj))
            for rec in result:
                if rec["reln"][0] == "_":  # skip _prev, _next, and convenience links
                    break
                ocls = None
                for lbl in rec["a"].labels:
                    ocls = self.cls_by_label(lbl)
                    if ocls:
                        break
                assert ocls
                o = ocls(rec["a"])
                # creating object but not putting in the cache - why?
                keys = self.keys_by_cls_and_reln(type(o), rec["reln"])
                # obj.belongs[(id(o),*keys)] = o
                # not setting the belongs on the obj - why?
                ret.append((o, keys))
        return ret

    def get_q(self, obj):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        return "MATCH (n:{lbl}) WHERE id(n)={neoid} RETURN n,id(n)".format(
            lbl=self.cls.mapspec()["label"], neoid=obj.neoid
        )

    def get_by_id_q(self):
        return "MATCH (n:{lbl}) WHERE id(n)=$id and n._to IS NULL RETURN id(n)".format(
            lbl=self.cls.mapspec()["label"]
        )

    def get_by_node_nanoid_q(self):
        """PROTOTYPE"""
        return "MATCH (n:node) WHERE n.nanoid=$nanoid and n._to is NULL RETURN id(n)"

    def get_attr_q(self, obj, att):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            return ""
        label = self.cls.mapspec()["label"]
        if att in self.cls.mapspec()["property"]:
            pr = self.cls.mapspec()["property"][att]
            return "MATCH (n:{lbl}) WHERE id(n)={neoid} RETURN n.{pr}".format(
                lbl=label, neoid=obj.neoid, pr=pr
            )
        elif att in self.cls.mapspec()["relationship"]:
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[\2]-\3", spec["rel"])
            if len(end_lbls) == 1:
                qry = "MATCH (n:{llbl}){rel}(a:{rlbl}) WHERE id(n)={neoid} RETURN a".format(
                    neoid=obj.neoid, llbl=label, rel=rel, rlbl=end_lbls[0]
                )
                if self.cls.attspec[att] == "object":
                    qry += " LIMIT 1"
                return qry
            else:  # multiple end classes possible
                cond = []
                for l in end_lbls:
                    cond.append("'{lbl}' IN labels(a)".format(lbl=l))
                cond = " OR ".join(cond)
                return "MATCH (n:{lbl}){rel}(a) WHERE id(n)={neoid} AND ({cond}) RETURN a".format(
                    lbl=label, rel=rel, neoid=obj.neoid, cond=cond
                )
        else:
            raise ArgError(
                "'{att}' is not a registered attribute for class '{cls}'".format(
                    att=att, cls=self.cls.__name__
                )
            )

    def get_owners_q(self, obj):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        label = self.cls.mapspec()["label"]
        return "MATCH (n:{lbl})<-[r]-(a) WHERE id(n)={neoid} RETURN TYPE(r) as reln, a".format(
            neoid=obj.neoid, lbl=label
        )

    def put_q(self, obj):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        props = {}
        null_props = []
        for pr in self.cls.mapspec()["property"]:
            if getattr(obj, pr) is None:
                null_props.append(self.cls.mapspec()["property"][pr])
            else:
                props[self.cls.mapspec()["property"][pr]] = getattr(obj, pr)
        stmts = []
        if obj.neoid is not None:
            set_clause = []
            for pr in props:
                set_clause.append(
                    "n.{pr}={val}".format(pr=pr, val=ObjectMap._quote_val(props[pr]))
                )
            set_clause = "SET " + ",".join(set_clause)
            stmts.append(
                "MATCH (n:{lbl}) WHERE id(n)={neoid} {set_clause} RETURN n,id(n)".format(
                    lbl=self.cls.mapspec()["label"],
                    neoid=obj.neoid,
                    set_clause=set_clause,
                )
            )
            for pr in null_props:
                stmts.append(
                    "MATCH (n:{lbl}) WHERE id(n)={neoid} REMOVE n.{pr} RETURN n,id(n)".format(
                        lbl=self.cls.mapspec()["label"], neoid=obj.neoid, pr=pr
                    )
                )
            return stmts
        else:
            spec = []
            for pr in props:
                spec.append(
                    "{pr}:{val}".format(pr=pr, val=ObjectMap._quote_val(props[pr]))
                )
            spec = ",".join(spec)
            return [
                "CREATE (n:%s {%s}) RETURN n,id(n)"
                % (self.cls.mapspec()["label"], spec)
            ]

    def put_attr_q(self, obj, att, values):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        if not isinstance(values, (Entity, list, CollValue)):
            raise ArgError(
                "'values' must be a list of values suitable for the attribute"
            )
        if isinstance(values, CollValue):
            values = values.values()
        elif isinstance(values, Entity):
            values = [values]
        if att in self.cls.mapspec()["property"]:
            return "MATCH (n:{lbl}) WHERE id(n)={neoid} SET {pr}={val} RETURN id(n)".format(
                lbl=self.cls.mapspec()["label"],
                neoid=obj.neoid,
                pr=self.cls.mapspec()["property"][att],
                val=ObjectMap._quote_val(values[0]),
            )
        elif att in self.cls.mapspec()["relationship"]:
            if not self._check_values_list(att, values):
                raise ArgError(
                    "'values' must be a list of mapped Entity objects of "
                    "the appropriate subclass for attribute '{att}'".format(
                        att=att
                    )
                )
            stmts = []
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[\2]-\3", spec["rel"])
            cond = []
            for l in end_lbls:
                cond.append("'{lbl}' IN labels(a)".format(lbl=l))
            cond = " OR ".join(cond)
            for avalue in values:
                if len(end_lbls) == 1:
                    stmts.append(
                        "MATCH (n:{lbl}),(a:{albl}) WHERE id(n)={neoid} AND id(a)={aneoid} "
                        "MERGE (n){rel}(a) RETURN id(a)".format(
                            lbl=self.cls.mapspec()["label"],
                            albl=end_lbls[0],
                            neoid=obj.neoid,
                            aneoid=avalue.neoid,
                            rel=rel,
                        )
                    )
                else:
                    stmts.append(
                        "MATCH (n:{lbl}),(a) WHERE id(n)={neoid} AND id(a)={aneoid} AND ({cond}) "
                        "MERGE (n){rel}(a) RETURN id(a)".format(
                            lbl=self.cls.mapspec()["label"],
                            cond=cond,
                            neoid=obj.neoid,
                            aneoid=avalue.neoid,
                            rel=rel,
                        )
                    )
            return stmts

        else:
            raise ArgError(
                "'{att}' is not a registered attribute for class '{cls}'".format(
                    att=att, cls=self.cls.__name__
                )
            )

    def rm_q(self, obj, detach=False):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        dlt = "DETACH DELETE n" if detach else "DELETE n"
        qry = "MATCH (n:{lbl}) WHERE id(n)={neoid} ".format(
            lbl=self.cls.mapspec()["label"], neoid=obj.neoid
        )
        return qry + dlt
        pass

    def rm_attr_q(self, obj, att, values=None):
        if not isinstance(obj, self.cls):
            raise ArgError(
                "arg1 must be object of class {cls}".format(cls=self.cls.__name__)
            )
        if obj.neoid is None:
            raise ArgError("object must be mapped (i.e., obj.neoid must be set)")
        if values and not isinstance(values, list):
            values = [values]
        if att in self.cls.mapspec()["property"]:
            return "MATCH (n:{lbl}) WHERE id(n)={neoid} REMOVE n.{att} RETURN id(n)".format(
                lbl=self.cls.mapspec()["label"], neoid=obj.neoid, att=att
            )
        elif att in self.cls.mapspec()["relationship"]:
            many = self.cls.attspec[att] == "collection"
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            cond = []
            for l in end_lbls:
                cond.append("'{lbl}' IN labels(a)".format(lbl=l))
            cond = " OR ".join(cond)
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[r\2]-\3", spec["rel"])
            if values[0] == ":all":
                if len(end_lbls) == 1:
                    return "MATCH (n:{lbl}){rel}(a:{albl}) WHERE id(n)={neoid} DELETE r RETURN id(n),id(a)".format(
                        lbl=self.cls.mapspec()["label"],
                        albl=end_lbls[0],
                        rel=rel,
                        neoid=obj.neoid,
                    )
                else:
                    return "MATCH (n:{lbl}){rel}(a) WHERE id(n)={neoid} AND ({cond}) DELETE r RETURN id(n)".format(
                        lbl=self.cls.mapspec()["label"],
                        cond=cond,
                        neoid=obj.neoid,
                        rel=rel,
                    )
            else:
                stmts = []

                if not self._check_values_list(att, values):
                    raise ArgError(
                        "'values' must be a list of mapped Entity objects of the "
                        "appropriate subclass for attribute '{att}'".format(att=att)
                    )
                for val in values:
                    qry = ""
                    if len(end_lbls) == 1:
                        qry = "MATCH (n:{lbl}){rel}(a:{albl}) WHERE id(n)={neoid} AND id(a)={aneoid} " \
                              "DELETE r RETURN id(n),id(a)".format(
                                  lbl=self.cls.mapspec()["label"],
                                  albl=end_lbls[0],
                                  neoid=obj.neoid,
                                  aneoid=val.neoid,
                                  rel=rel)
                    else:
                        qry = "MATCH (n:{lbl}){rel}(a) WHERE id(n)={neoid} AND id(a)={aneoid} AND ({cond}) " \
                              "DELETE r RETURN id(n),id(a)".format(
                                  lbl=self.cls.mapspec()["label"],
                                  neoid=obj.neoid,
                                  aneoid=val.neoid,
                                  cond=cond,
                                  rel=rel)
                    stmts.append(qry)
                return stmts
        else:
            raise ArgError(
                "'{att}' is not a registered attribute for class '{cls}'".format(
                    att=att, cls=self.cls.__name__
                )
            )

    def _check_values_list(self, att, values):
        v = values
        if isinstance(values, CollValue):
            v = values.values()
        chk = [x.neoid is None for x in v]
        if True in chk:
            return False
        end_cls = self.cls.mapspec()["relationship"][att]["end_cls"]
        if isinstance(end_cls, str):
            end_cls = {end_cls}
        cls_set = tuple([eval(x) for x in end_cls])
        chk = [isinstance(x, cls_set) for x in v]
        if True in chk:
            return True
        return False



import fastavro
from fastavro.schema import load_schema
from fastavro.validation import validate


from tempfile import NamedTemporaryFile
from pdb import set_trace
import json

# the following gyrations enable
# * the PFB schema to be modularized with named types,
# * the addition the custom data types to the pfb.Entity schema, and
# * the recursive loading of the named type schemas by fastavro.

pfb_schema = None
with open("pfb.Entity.avsc","r") as Entity:
    # load Entity schema as simple json
    pfb_schema_json = json.load( Entity )
    # find the "object" hash 
    [object] = [ x for x in pfb_schema_json["fields"] if x["name"] == "object" ]
    # add the custom schemas (as names) to the object.type array
    object["type"].extend([ "icdc.case", "icdc.cohort" ])
    # dump json to a tempfile to take advantage of fastavro avsc 
    # name resolution in fastavro.schema.load_schema()
    tf = NamedTemporaryFile(mode="w+",dir=".")
    json.dump(pfb_schema_json,tf)
    tf.seek(0)
    # load the customized schema
    pfb_schema = load_schema(tf.name)
    pass

# metadata for PFB message:
# note that unused fields still must be defined, since in the schema
# null is not allowed (which seems more like oversight than intention)

icdc_cohort_meta = { 
    "name": "icdc.cohort",
    "ontology_reference": "",
    "values": {},
    "links":[],
    "properties": [
        { 
            "name": "cohort_description",
            "ontology_reference": "NCIT",
            "values": {
                "concept_code": "C166209"
                }
        },
        { 
            "name": "cohort_dose",
            "ontology_reference": "NCIT",
            "values": {
                "concept_code": "C166210"
                }
        }
    ]
}

icdc_case_meta = {
    "name": "icdc.case",
    "ontology_reference": "",
    "values": {},
    "properties": [
        { 
            "name": "case_id",
            "ontology_reference": "NCIT",
            "values": {
              "concept_code": "C164324"
            }
        },
        { 
            "name": "patient_id",
            "ontology_reference": "NCIT",
            "values": {
              "concept_code": "C164337"
            }
        }
        ],
        "links": [
            {
                "name": "member_of",
                "dst": "cohort",
                "multiplicity": "MANY_TO_ONE"
            }
        ]
}

# Metadata    
assert validate( {
    "name":"Metadata",
    "misc":{},
    "nodes": [
        icdc_case_meta,
        icdc_cohort_meta
    ]}, pfb_schema)


# data for PFB message:

cohort_data = {
    "id": "n201",
    "cohort_description": "arm1",
    "cohort_dose": "10mg/kg"
}

assert validate( ("icdc.cohort", cohort_data), pfb_schema )

case_data = {
    "id": "n101",
    "case_id": "UBC01-007",
    "patient_id": "007",
    "patient_first_name": "Fluffy"
}    

assert validate( ("icdc.case", case_data), pfb_schema )

link = {
    "dst_name": "icdc.cohort",
    "dst_id": "n201"
}


case_data_entity = {
    "name": "icdc.case",
    "id": "n101",
    "object": case_data,
    "relations": [ link ]
}

assert validate( ("pfb.Entity", case_data_entity), pfb_schema )

cohort_data_entity = {
    "name":"icdc.cohort",
    "id": "n201",
    "object": cohort_data,
    "relations":[]
}

assert validate( ("pfb.Entity", cohort_data_entity), pfb_schema )

payload = [
      { 
        "name": "Metadata",
        "object": {
            "name": "pfb.Metadata",
            "misc": {},
            "nodes": [
                icdc_cohort_meta,
                icdc_case_meta
            ]
        }
      },
      cohort_data_entity,
      case_data_entity
    ]    

# Create PFB message

with open("worked-example.avro","wb") as out:
    fastavro.writer( out, pfb_schema, payload)

# Read records from message

with open("worked-example.avro","rb") as inf:
    rdr = fastavro.reader(inf)
    for rec in rdr:
        print(rec)




"""
makeq - make a Neo4j query from an endpoint path.
"""
import yaml
import re
from pdb import set_trace
from bento_meta.util._engine import _engine
from bento_meta.util.cypher.entities import (  # noqa E402
    N, R, P, N0, R0, G,
    _as, _var, _plain, _anon,
    )
from bento_meta.util.cypher.functions import (
    count, exists, labels, group, And, Or, Not,
    )
from bento_meta.util.cypher.clauses import (
    Match, Where, With, Return,
    Statement,
    )

avail_funcs = {x.__name__: x for x in (count, exists, labels, group, And, Or, Not)}


def f(pfx, pth):
    tok = [x for x in pth if x.startswith('$')]
    if not tok:
        tok = [x for x in pth if not x.startswith('_')]
    if not tok:
        print(pfx)
        return
    else:
        if pth.get('_return'):
            print(pfx)
        for t in tok:
            f('/'.join([pfx, t]), pth[t])
        return

class Query(object):
    paths = {}
    cache = {}

    def __init__(self, path, use_cache=True):
        if path.startswith("/"):
            path = path[1:]
        self.toks = path.split("/")
        self._engine = None
        if use_cache:
            # interpret the cache key as a regexp matching the input path
            hit = [x for x in self.cache if re.match("^"+x+"$", path)]
            if hit:
                Q = self.cache[hit[0]]
                # pull the new parameter values from the path
                vals = re.match(hit[0], path).groups()
                keys = sorted(Q._engine.params.keys())
                self._engine = Q._engine
                for pr in zip(keys, vals):
                    self._engine.params[pr[0]] = pr[1]
        if not self._engine:
            self._engine = _engine()
            if not self._engine.parse(self.toks):
                raise RuntimeError(self._engine.error)
            if use_cache:
                self.cache[self._engine.key] = self

    @classmethod
    def set_paths(cls, paths):
        if paths.get('paths'):
            cls.paths = paths['paths']
        else:
            cls.paths = paths
        _engine.set_paths(cls.paths)
        return True

    @classmethod
    def load_paths(cls, flo):
        p = yaml.load(flo, Loader=yaml.CLoader)
        return cls.set_paths(p)


    @property
    def statement(self):
        return self._engine.statement

    @property
    def params(self):
        return self._engine.params

    @property
    def path_id(self):
        return self._engine.path_id
    
    def __str__(self):
        return str(self.statement)

